# -*- coding: utf-8 -*-

"""Implementation of the canonicalization algorithm."""

from operator import attrgetter
from typing import Collection, Iterable, Mapping, Optional, Sequence, Tuple, Union

from ..dsl import (
    CounterfactualVariable,
    Distribution,
    Expression,
    Fraction,
    One,
    Probability,
    Product,
    Sum,
    Variable,
    Zero,
    ensure_ordering,
)

__all__ = [
    "canonicalize",
    "canonical_expr_equal",
]


def canonicalize(
    expression: Expression, ordering: Optional[Sequence[Union[str, Variable]]] = None
) -> Expression:
    """Canonicalize an expression that meets the markov condition with respect to the given ordering.

    :param expression: An expression to canonicalize
    :param ordering: A toplogical ordering. If none is given, it is assigned by sort order of the variable names.
    :return: A canonical expression
    """
    canonicalizer = Canonicalizer(ensure_ordering(expression, ordering=ordering))
    return canonicalizer.canonicalize(expression)


class Canonicalizer:
    """A data structure to support application of the canonicalize algorithm."""

    ordering: Sequence[Variable]
    ordering_level: Mapping[str, int]

    def __init__(self, ordering: Sequence[Variable]) -> None:
        """Initialize the canonicalizer.

        :param ordering: A topological ordering over the variables appearing in the expression.

        :raises ValueError: if the ordering has duplicates
        """
        if len(set(ordering)) != len(ordering):
            raise ValueError(f"ordering has duplicates: {ordering}")

        self.ordering = ordering
        self.ordering_level = {variable.name: level for level, variable in enumerate(self.ordering)}

    def _canonicalize_probability(self, expression: Probability) -> Probability:
        return expression._new(
            Distribution(
                children=self._sorted(expression.children),
                parents=self._sorted(expression.parents),
            )
        )

    def _sorted(self, variables: Collection[Variable]) -> Tuple[Variable, ...]:
        return tuple(
            sorted(
                (self._canonicalize_variable(variable) for variable in variables),
                key=self._sorted_key,
            )
        )

    def _canonicalize_variable(self, variable: Variable) -> Variable:
        if isinstance(variable, CounterfactualVariable):
            return CounterfactualVariable(
                name=variable.name,
                star=variable.star,
                interventions=tuple(sorted(variable.interventions)),
            )
        else:
            return variable

    def _sorted_key(self, variable: Variable) -> int:
        return self.ordering_level[variable.name]

    def canonicalize(self, expression: Expression) -> Expression:
        """Canonicalize an expression.

        :param expression: An uncanonicalized expression
        :return: A canonicalized expression
        :raises TypeError: if an object with an invalid type is passed
        """
        if isinstance(expression, Probability):  # atomic
            return self._canonicalize_probability(expression)
        elif isinstance(expression, Sum):
            return Sum.safe(
                expression=self.canonicalize(expression.expression),
                ranges=expression.ranges,
                simplify=True,
            )
        elif isinstance(expression, Product):
            # note: safe already sorts
            return Product.safe(
                self.canonicalize(subexpr) for subexpr in _flatten_product(expression)
            )
        elif isinstance(expression, Fraction):
            numerator = self.canonicalize(expression.numerator)
            # TODO check if there's a zero in numerator, then return zero if so
            denominator = self.canonicalize(expression.denominator)
            if isinstance(denominator, One):
                return numerator
            if numerator == denominator:
                return One()
            return numerator / denominator  # TODO
        elif isinstance(expression, (One, Zero)):
            return expression
        else:
            raise TypeError


def _flatten_product(product: Product) -> Iterable[Expression]:
    for expression in product.expressions:
        if isinstance(expression, Product):
            yield from _flatten_product(expression)
        else:
            yield expression


def canonical_expr_equal(left: Expression, right: Expression) -> bool:
    """Return True if two expressions are equal after canonicalization."""
    ordering = sorted(left.get_variables() | right.get_variables(), key=attrgetter("name"))
    return canonicalize(left, ordering) == canonicalize(right, ordering)



# -*- coding: utf-8 -*-

"""Implementation of the IDC algorithm."""

from .id_std import identify
from .utils import Identification
from ..conditional_independencies import are_d_separated
from ...dsl import Expression, Variable

__all__ = [
    "idc",
    "rule_2_of_do_calculus_applies",
]


def idc(identification: Identification) -> Expression:
    """Run the IDC algorithm.

    :param identification: The identification tuple
    :returns: An expression created by the :func:`identify` algorithm after simplifying the original query
    """
    for condition in identification.conditions:
        if rule_2_of_do_calculus_applies(identification=identification, condition=condition):
            return idc(identification.exchange_observation_with_action(condition))

    # Run ID algorithm
    id_estimand = identify(identification.uncondition())
    return id_estimand.normalize_marginalize(identification.outcomes)


def rule_2_of_do_calculus_applies(identification: Identification, condition: Variable) -> bool:
    r"""Check if Rule 2 of the Do-Calculus applies to the conditioned variable.

    :param identification: The identification tuple
    :param condition: The condition to check
    :returns: If rule 2 applies, see below.

    If Rule 2 of the do calculus applies to the conditioned variable, then it can be converted to a do variable.

    .. math::

        \newcommand\ci{\perp\!\!\!\perp}
        \newcommand{\ubar}[1]{\underset{\bar{}}{#1}}
        \newcommand{\obar}[1]{\overset{\bar{}}{#1}}
        \text{if } (\exists Z \in \mathbf{Z})(\mathbf{Y} \ci Z | \mathbf{X}, \mathbf{Z}
        - \{Z\})_{G_{\bar{\mathbf{X}}\ubar{Z}}} \\
        \text{then } P(\mathbf{Y}|do(\mathbf{X}),\mathbf{Z}) = P(\mathbf Y|do(\mathbf X), do(Z), \mathbf{Z} - \{Z\})
    """
    graph = identification.graph
    treatments = identification.treatments
    conditions = treatments | (identification.conditions - {condition})
    graph_mod = graph.remove_in_edges(treatments).remove_out_edges(condition)
    return all(
        are_d_separated(graph_mod, outcome, condition, conditions=conditions)
        for outcome in identification.outcomes
    )



# -*- coding: utf-8 -*-

"""Identification algorithms."""

from .id_c import idc
from .id_star import id_star
from .id_std import identify, identify_outcomes
from .idc_star import idc_star
from .utils import Identification, Query, Unidentifiable

__all__ = [
    # Algorithms
    "identify",
    "identify_outcomes",
    "id_star",
    "idc",
    "idc_star",
    # Data Structures
    "Query",
    # Exceptions
    "Unidentifiable",
    "Identification",
]



# -*- coding: utf-8 -*-

"""Causal graphs have implications that can be tested in the context of a specific dataset.

This module includes algorithms to perform those tests.
"""

from dataclasses import dataclass
from typing import Iterable, Optional, Union

import pandas as pd
from tqdm.auto import tqdm

from .conditional_independencies import get_conditional_independencies
from ..graph import NxMixedGraph
from ..struct import DSeparationJudgement

__all__ = [
    "get_graph_falsifications",
    "get_falsifications",
    "Falsifications",
]


@dataclass
class Falsifications:
    """A list of variables pairs that failed the D-separation and covariance test.

    Has an extra 'evidence' property that is a dictionary.

    - Keys are the d-separated variable pairs
    - Values are the covariances measured between them.
    """

    #: Sequence of implications that did not pass
    failures: pd.Series
    #: Collection of all implications tested
    evidence: pd.DataFrame


def get_graph_falsifications(
    graph: NxMixedGraph,
    df: pd.DataFrame,
    *,
    significance_level: Optional[float] = None,
    max_given: Optional[int] = None,
    verbose: bool = False,
    method: Optional[str] = None,
) -> Falsifications:
    """Test conditional independencies implied by a graph.

    :param graph: An ADMG
    :param df: Data to check for consistency with a causal implications
    :param significance_level: Significance for p-value test
    :param max_given: The maximum set size in the power set of the vertices minus the d-separable pairs
    :param verbose: If true, use tqdm for status updates.
    :param method: Conditional independence from :mod:`pgmpy` to use. If none,
        defaults to :func:`pgmpy.estimators.CITests.cressie_read`.
    :return: Falsifications report
    """
    judgements = get_conditional_independencies(graph, max_conditions=max_given, verbose=verbose)
    return get_falsifications(
        judgements=judgements,
        df=df,
        significance_level=significance_level,
        verbose=verbose,
        method=method,
    )


HB_LEVEL_NAME = "Holm–Bonferroni level"


def get_falsifications(
    judgements: Union[NxMixedGraph, Iterable[DSeparationJudgement]],
    df: pd.DataFrame,
    *,
    significance_level: Optional[float] = None,
    verbose: bool = False,
    method: Optional[str] = None,
) -> Falsifications:
    """Test conditional independencies implied by a list of D-separation judgements.

    :param judgements: A list of D-separation judgements to check.
    :param df: Data to check for consistency with a causal implications
    :param significance_level: Significance for p-value test
    :param verbose: If true, use tqdm for status updates.
    :param method: Conditional independence from :mod:`pgmpy` to use. If none,
        defaults to :func:`pgmpy.estimators.CITests.cressie_read`.
    :return: Falsifications report
    """
    if significance_level is None:
        significance_level = 0.05
    variances = {
        judgement: judgement.test(df, test=method)
        for judgement in tqdm(judgements, disable=not verbose, desc="Checking conditionals")
    }
    evidence = pd.DataFrame(
        [
            (
                judgement.left.name,
                judgement.right.name,
                "|".join(c.name for c in judgement.conditions),
                chi,
                p,
                dof,
            )
            for judgement, (chi, p, dof) in variances.items()
        ],
        columns=["left", "right", "given", "chi^2", "p", "dof"],
    )
    evidence.sort_values("p", ascending=True, inplace=True)
    evidence = (
        evidence.assign(
            **{HB_LEVEL_NAME: significance_level / pd.Series(range(len(evidence.index) + 1, 0, -1))}
        )
        .pipe(_assign_flags)
        .sort_values(["flagged", "dof"], ascending=False)
    )

    failures_df = evidence.loc[evidence["flagged"], ["left", "right", "given"]]
    failures = failures_df.apply(tuple, axis="columns")
    return Falsifications(failures, evidence)


def _assign_flags(df: pd.DataFrame) -> pd.DataFrame:
    return df.assign(flagged=(df["p"] < df[HB_LEVEL_NAME]))



# -*- coding: utf-8 -*-

"""Tests for expression complexity.

Note: tests in this module should **NOT** use exact
numbers, but should rather encode our intuitive ideas
of what's more complex by comparing expressions.
"""

import unittest

from y0.complexity import complexity
from y0.dsl import A, B, C, D, Expression, Fraction, One, P, X, Y, Z
from y0.mutate import bayes_expand, chain_expand, fraction_expand


class TestComplexity(unittest.TestCase):
    """Test case for complexity tests."""

    def assert_complexity_le(self, left: Expression, right: Expression) -> None:
        """Assert that the complexity of the first expression is less than the second."""
        self.assertIsInstance(left, Expression)
        self.assertIsInstance(right, Expression)
        self.assertLessEqual(complexity(left), complexity(right))

    def assert_complexity_equal(self, left: Expression, right: Expression) -> None:
        """Assert that the complexity of the first expression is less than the second."""
        self.assertIsInstance(left, Expression)
        self.assertIsInstance(right, Expression)
        self.assertEqual(complexity(left), complexity(right))

    def test_complexity_equal(self):
        """Test complexity equivalence."""
        examples = [
            (P(A), P(B), "probability 1-variable isomorphism"),
            (P(A, B), P(B, A), "probability 2-variable isomorphism, same variables"),
            (P(A, B), P(B, C), "probability 2-variable isomorphism, one different"),
            (P(A, B), P(C, D), "probability 2-variable isomorphism, two different"),
        ]
        for left, right, label in examples:
            with self.subTest(label=label):
                self.assert_complexity_equal(left, right)

    def test_expand(self):
        """Test various expansions always make at least more complicated expressions."""
        examples = [
            P(A | X),
            P(A | X, Y),
            P(A | X, Y, Z),
            P(A, B | X),
            P(A, B | X, Y),
            P(A, B | X, Y, Z),
        ]
        for example in examples:
            with self.subTest(expr=example.to_y0(), type="fraction"):
                self.assert_complexity_le(example, fraction_expand(example))
            with self.subTest(expr=example.to_y0(), type="bayes"):
                self.assert_complexity_le(example, bayes_expand(example))
            with self.subTest(expr=example.to_y0(), type="chain"):
                self.assert_complexity_le(example, chain_expand(example))

    def test_fraction_simplify(self):
        """Test simplifying a fraction always results in at least a less complicated expression."""
        examples = [
            Fraction(P(A), One()),
            Fraction(P(A) * P(B), P(B)),
        ]
        for example in examples:
            with self.subTest(expr=example.to_y0()):
                self.assertIsInstance(example, Fraction)
                self.assert_complexity_le(example.simplify(), example)



# -*- coding: utf-8 -*-

"""Tests for simplifying latent variable DAGs."""

import unittest

import networkx as nx

from y0.algorithm.simplify_latent import (
    DEFAULT_SUFFIX,
    iter_latents,
    remove_redundant_latents,
    remove_widow_latents,
    transform_latents_with_parents,
)
from y0.algorithm.taheri_design import taheri_design_dag
from y0.dsl import Y1, Y2, Y3, U, Variable, W
from y0.examples import igf_example
from y0.graph import set_latent

X1, X2, X3 = map(Variable, ["X1", "X2", "X3"])


class TestDesign(unittest.TestCase):
    """Test the design algorithm."""

    def test_design(self):
        """Test the design algorithm."""
        results = taheri_design_dag(igf_example.graph.directed, cause="PI3K", effect="Erk", stop=3)
        self.assertIsNotNone(results)
        # FIXME do better than this.


class TestSimplify(unittest.TestCase):
    """Tests for the Robin Evans simplification algorithms."""

    def assert_latent_variable_dag_equal(self, expected, actual) -> None:
        """Check two latent variable DAGs are the same."""
        self.assertEqual(
            sorted(expected),
            sorted(actual),
            msg="Nodes are incorrect",
        )
        self.assertEqual(
            dict(expected.nodes.items()),
            dict(actual.nodes.items()),
            msg="Tags are incorrect",
        )
        self.assertEqual(
            sorted(expected.edges()),
            sorted(actual.edges()),
        )

    def test_remove_widows(self):
        """Test simplification 1 - removing widows."""
        graph = nx.DiGraph()
        graph.add_edges_from(
            [
                (X1, X2),
                (X2, W),
                (U, X2),
                (U, X3),
                (U, W),
            ]
        )
        latents = {U, W}
        set_latent(graph, latents)
        self.assertEqual(latents, set(iter_latents(graph)))

        # Apply the simplification
        _, removed = remove_widow_latents(graph)
        self.assertEqual({W}, removed)

        expected = nx.DiGraph()
        expected.add_edges_from(
            [
                (X1, X2),
                (U, X2),
                (U, X3),
            ]
        )
        set_latent(expected, U)

        self.assert_latent_variable_dag_equal(expected, graph)

    def test_transform_latents_with_parents(self):
        """Test simplification 2: latents with parents can be transformed."""
        graph = nx.DiGraph()
        graph.add_edges_from(
            [
                (X1, U),
                (X2, U),
                (U, Y1),
                (U, Y2),
                (U, Y3),
            ]
        )
        set_latent(graph, U)
        # Apply the simplification
        transform_latents_with_parents(graph)

        expected = nx.DiGraph()
        expected.add_edges_from(
            [
                (X1, Y1),
                (X1, Y2),
                (X1, Y3),
                (X2, Y1),
                (X2, Y2),
                (X2, Y3),
                (Variable(f"U{DEFAULT_SUFFIX}"), Y1),
                (Variable(f"U{DEFAULT_SUFFIX}"), Y2),
                (Variable(f"U{DEFAULT_SUFFIX}"), Y3),
            ]
        )
        set_latent(expected, Variable(f"U{DEFAULT_SUFFIX}"))

        self.assert_latent_variable_dag_equal(expected, graph)

    def test_remove_redundant_latents(self):
        """Test simplification 3 - remove redundant latents."""
        graph = nx.DiGraph()
        graph.add_edges_from(
            [
                (U, X1),
                (U, X2),
                (U, X3),
                (W, X1),
                (W, X2),
            ]
        )
        set_latent(graph, [U, W])
        # Apply the simplification
        remove_redundant_latents(graph)

        expected = nx.DiGraph()
        expected.add_edges_from(
            [
                (U, X1),
                (U, X2),
                (U, X3),
            ]
        )
        set_latent(expected, [U])

        self.assert_latent_variable_dag_equal(expected, graph)



# -*- coding: utf-8 -*-
# type: ignore

"""Examples from CausalFusion."""

from __future__ import annotations

import numpy as np
import pandas as pd

from .backdoor import generate_data_for_backdoor
from .frontdoor import generate_data_for_frontdoor
from .frontdoor_backdoor import generate_data_for_frontdoor_backdoor
from .sars import generate_data_for_covid_case_study
from .smoke_cancer import generate_data_for_smoke_cancer
from .utils import Example
from ..algorithm.identify import Identification, Query
from ..dsl import (
    AA,
    W0,
    W1,
    W2,
    X1,
    X2,
    Y1,
    Y2,
    Z1,
    Z2,
    Z3,
    Z4,
    Z5,
    A,
    B,
    C,
    D,
    E,
    F,
    G,
    M,
    P,
    Q,
    S,
    Sum,
    T,
    Variable,
    W,
    X,
    Y,
    Z,
)
from ..graph import NxMixedGraph
from ..resources import ASIA_PATH
from ..struct import DSeparationJudgement, VermaConstraint

x, y, z, w = -X, -Y, -Z, -W

u_2 = Variable("u_2")
u_3 = Variable("u_3")

#: Treatment: X
#: Outcome: Y
#: Adjusted: N/A
backdoor = NxMixedGraph.from_edges(
    directed=[
        (Z, X),
        (Z, Y),
        (X, Y),
    ]
)

backdoor_example = Example(
    name="Backdoor",
    reference='J. Pearl. 2009. "Causality: Models, Reasoning and Inference.'
    ' 2nd ed." Cambridge University Press, p. 178.',
    graph=backdoor,
    generate_data=generate_data_for_backdoor,
    example_queries=[Query.from_str(treatments="X", outcomes="Y")],
)

#: Treatment: X
#: Outcome: Y
#: Adjusted: N/A
frontdoor = NxMixedGraph.from_edges(
    directed=[
        (X, Z),
        (Z, Y),
    ],
    undirected=[
        (X, Y),
    ],
)
frontdoor_example = Example(
    name="Frontdoor",
    reference='J. Pearl. 2009. "Causality: Models, Reasoning and Inference.'
    ' 2nd ed." Cambridge University Press, p. 81.',
    graph=frontdoor,
    generate_data=generate_data_for_frontdoor,
    example_queries=[Query.from_str(treatments="X", outcomes="Y")],
)

#: Treatment: X
#: Outcome: Y
#: Adjusted: N/A
frontdoor_backdoor = NxMixedGraph.from_edges(
    directed=[
        (X, Z),
        (Z, Y),
        (W, X),
        (W, Y),
    ],
)
frontdoor_backdoor_example = Example(
    name="Frontdoor / Backdoor",
    reference="https://github.com/y0-causal-inference/y0/pull/183",
    graph=frontdoor_backdoor,
    generate_data=generate_data_for_frontdoor_backdoor,
    example_queries=[Query.from_str(treatments="X", outcomes="Y")],
)

#: Treatment: X
#: Outcome: Y
instrumental_variable = NxMixedGraph.from_edges(
    directed=[
        (Z, X),
        (X, Y),
    ],
    undirected=[
        (X, Y),
    ],
)
instrumental_variable_example = Example(
    name="Instrument Variable",
    reference='J. Pearl. 2009. "Causality: Models, Reasoning and Inference.'
    ' 2nd ed." Cambridge University Press, p. 153.',
    graph=instrumental_variable,
)

#: Treatment: X
#: Outcome: Y
napkin = NxMixedGraph.from_edges(
    directed=[
        (Z2, Z1),
        (Z1, X),
        (X, Y),
    ],
    undirected=[
        (Z2, X),
        (Z2, Y),
    ],
)


def generate_napkin_data(
    num_samples: int, treatments: dict[Variable, float] | None = None, *, seed: int | None = None
) -> pd.DataFrame:
    """Generate testing data for the napkin graph.

    :param num_samples: The number of samples to generate. Try 1000.
    :param treatments: An optional dictionary of the values to fix each variable
        to. The keys in this dictionary must correspond to variables in the
        napkin graph as defined in :data:`y0.examples.napkin` (i.e.,
        with :data:`y0.dsl.Z1`, :data:`y0.dsl.Z2`, :data:`y0.dsl.X`,
        and :data:`y0.dsl.Y`).
    :param seed: An optional random seed for reproducibility purposes
    :returns: A pandas Dataframe with columns corresponding to the four
        variable names in the Napkin graph (i.e., ``Z1``, ``Z2``, ``X``,
        and ``Y``)

    Generate _observational_ data with the following:

    >>> from y0.examples.napkin_example
    >>> napkin_example.generate_data(1000)

    Generate interventional data on $X=1$ with the following:

    >>> from y0.dsl import X
    >>> napkin_example.generate_data(1000, treatments={X: 1})

    Multiple treatments can be specified:

    >>> from y0.dsl import X, Z1
    >>> napkin_example.generate_data(1000, treatments={X: 1, Z1: 0})
    """
    if treatments is None:
        treatments = {}
    generator = np.random.default_rng(seed)
    # U1 is the latent variable that is a common cause of W and X
    u1 = generator.normal(loc=3, scale=1, size=num_samples)
    # U2 is the latent variable that is a common cause of W and Y
    u2 = generator.normal(loc=5, scale=1, size=num_samples)
    if Z2 in treatments:
        z2 = np.full(num_samples, treatments[Z2])
    else:
        u_linear_combination = 0.3 * u1 + 0.5 * u2
        z2 = generator.gamma(
            shape=u_linear_combination**-2,
            scale=5 * u_linear_combination,
            size=num_samples,
        )
    if Z1 in treatments:
        z1 = np.full(num_samples, treatments[Z1])
    else:
        z1 = generator.normal(loc=z2 * 0.7, scale=6, size=num_samples)
    if X in treatments:
        x = np.full(num_samples, treatments[X])
    else:
        x = generator.binomial(n=1, p=1 / (1 + np.exp(-2 - 0.23 * u1 - 0.1 * z1)), size=num_samples)
    if Y in treatments:
        y = np.full(num_samples, treatments[Y])
    else:
        y = generator.normal(loc=u2 * 0.5 + x * 3, scale=6)
    return pd.DataFrame({Z2.name: z2, Z1.name: z1, X.name: x, Y.name: y})


napkin_example = Example(
    name="Napkin",
    reference='J. Pearl and D. Mackenzie. 2018. "The Book of Why: The New Science of Cause and Effect."'
    " Basic Books, p. 240.",
    graph=napkin,
    generate_data=generate_napkin_data,
    example_queries=[Query.from_str(treatments="X", outcomes="Y")],
    verma_constraints=[
        VermaConstraint(
            lhs_cfactor=Q[X, Y](Z1, X, Y) / Sum[Y](Q[X, Y](Z1, X, Y)),
            lhs_expr=(
                Sum[Z2](P(Y | (Z1, Z2, X)) * P(X | (Z2, Z1)) * P(Z2))
                / Sum[Z2, Y](P(Y | (Z2, Z1, X)) * P(X | (Z2, Z1)) * P(Z2))
            ),
            rhs_cfactor=Q[Y](X, Y),
            rhs_expr=Sum[u_2, X](P(Y | u_2 | X) * P(X) * P(u_2)),
            variables=(Z1,),
        ),
    ],
)

#: Treatment: X
#: Outcome: Y
#: Reference:
m_graph = NxMixedGraph.from_edges(
    directed=[
        (X, Y),
    ],
    undirected=[
        (X, Z),
        (Y, Z),
    ],
)
m_graph_example = Example(
    name="M-Graph",
    reference='S. Greenland, J. Pearl, and J.M. Robins. 1999. "Causal Diagrams for Epidemiologic Research."'
    " Epidemiology Journal, Volume 10, No. 10, pp. 37-48, 1999.",
    graph=m_graph,
)

# NxMixedGraph containing vertices without edges
vertices_without_edges = Example(
    name="Vertices-without-Edges",
    reference="out of the mind of JZ (patent pending). See NFT for details",
    graph=NxMixedGraph.from_adj(
        directed={W: [], X: [Y], Y: [Z], Z: []},
        undirected={W: [], X: [Z], Y: [], Z: []},
    ),
)

# Line 1 example
line_1_example = Example(
    name="Line 1 of ID algorithm",
    reference="out of the mind of JZ",
    graph=NxMixedGraph.from_edges(
        directed=[
            (Z, Y),
        ]
    ),
    identifications=[
        dict(
            id_in=[
                Identification.from_expression(
                    query=P(Y),
                    estimand=P(Y, Z),
                    graph=NxMixedGraph.from_edges(directed=[(Z, Y)]),
                )
            ],
            id_out=[
                Identification.from_expression(
                    query=P(Y),
                    estimand=Sum[Z](P(Y, Z)),
                    graph=NxMixedGraph.from_edges(directed=[(Z, Y)]),
                )
            ],
        ),
        dict(
            id_in=[
                Identification.from_expression(
                    query=P(Y, Z),
                    estimand=P(Y, Z),
                    graph=NxMixedGraph.from_edges(directed=[(Z, Y)]),
                )
            ],
            id_out=[
                Identification.from_expression(
                    query=P(Y, Z),
                    estimand=P(Y, Z),
                    graph=NxMixedGraph.from_edges(directed=[(Z, Y)]),
                )
            ],
        ),
    ],
)

# Line 2 example
line_2_example = Example(
    name="intervention not ancestral to outcome",
    reference="out of the mind of JZ",
    graph=NxMixedGraph.from_edges(directed=[(Z, Y), (Y, X)], undirected=[(Z, X)]),
    identifications=[
        dict(
            id_in=[
                Identification.from_expression(
                    query=P(Y @ X),
                    estimand=P(X, Y, Z),
                    graph=NxMixedGraph.from_edges(directed=[(Z, Y), (Y, X)], undirected=[(Z, X)]),
                )
            ],
            id_out=[
                Identification.from_expression(
                    query=P(Y),
                    estimand=Sum[X](P(Y, X, Z)),
                    graph=NxMixedGraph.from_edges(directed=[(Z, Y)]),
                )
            ],
        )
    ],
)

line_3_example = Example(
    name="node has no effect on outcome",
    reference="out of the mind of JZ",
    graph=NxMixedGraph.from_edges(directed=[(Z, X), (X, Y)], undirected=[(Z, X)]),
    identifications=[
        dict(
            id_in=[
                Identification.from_expression(
                    query=P(Y @ X),
                    estimand=P(X, Y, Z),
                    graph=NxMixedGraph.from_edges(directed=[(Z, X), (X, Y)], undirected=[(Z, X)]),
                )
            ],
            id_out=[
                Identification.from_expression(
                    query=P(Y @ {X, Z}),
                    estimand=P(X, Y, Z),
                    graph=NxMixedGraph.from_edges(directed=[(Z, X), (X, Y)], undirected=[(Z, X)]),
                )
            ],
        ),
    ],
)

line_4_example = Example(
    name="graph without X decomposes into multiple C components",
    reference="out of the mind of JZ",
    graph=NxMixedGraph.from_edges(
        directed=[(X, M), (Z, X), (Z, Y), (M, Y)],
        undirected=[(Z, X), (M, Y)],
    ),
    identifications=[
        dict(
            id_in=[
                Identification.from_expression(
                    query=P(Y @ X),
                    estimand=P(M, X, Y, Z),
                    graph=NxMixedGraph.from_edges(
                        directed=[(X, M), (Z, X), (Z, Y), (M, Y)],
                        undirected=[(Z, X), (M, Y)],
                    ),
                )
            ],
            id_out=[
                Identification.from_expression(
                    query=P(M @ {X, Z}, Y @ {X, Z}),
                    estimand=P(M, X, Y, Z),
                    graph=NxMixedGraph.from_edges(
                        directed=[(X, M), (Z, X), (Z, Y), (M, Y)],
                        undirected=[(Z, X), (M, Y)],
                    ),
                ),
                Identification.from_expression(
                    query=P(Z @ {M, X, Y}),
                    estimand=P(M, X, Y, Z),
                    graph=NxMixedGraph.from_edges(
                        directed=[(X, M), (Z, X), (Z, Y), (M, Y)],
                        undirected=[(Z, X), (M, Y)],
                    ),
                ),
            ],
        ),
    ],
)

line_5_example = Example(
    name="graph containing a hedge",
    reference="Shpitser, I., & Pearl, J. (2008). Complete Identification Methods for the Causal Hierarchy.",
    graph=NxMixedGraph.from_edges(directed=[(X, Y)], undirected=[(X, Y)]),
    identifications=[
        dict(
            id_in=[
                Identification.from_expression(
                    query=P(Y @ X),
                    estimand=P(X, Y),
                    graph=NxMixedGraph.from_edges(directed=[(X, Y)], undirected=[(X, Y)]),
                )
            ],
        )
    ],
)

line_6_example = Example(
    name="ID Line 6 Example",
    description="If there are no bidirected arcs from X to the other nodes in the"
    " current subproblem under consideration, then we can replace acting"
    " on X by conditioning, and thus solve the subproblem.",
    reference="Shpitser, I., & Pearl, J. (2008). Complete Identification Methods for the Causal Hierarchy.",
    graph=NxMixedGraph.from_edges(directed=[(X, Y), (X, Z), (Z, Y)], undirected=[(X, Z)]),
    identifications=[
        dict(
            id_in=[
                Identification.from_expression(
                    query=P(Y @ [X, Z]),
                    estimand=P(X, Y, Z),
                    graph=NxMixedGraph.from_edges(
                        directed=[(X, Y), (X, Z), (Z, Y)],
                        undirected=[(X, Z)],
                    ),
                )
            ],
            id_out=[
                Identification.from_expression(
                    query=P(Y @ {X, Z}),
                    estimand=P(Y | [X, Z]),
                    graph=NxMixedGraph.from_edges(
                        directed=[(X, Y), (X, Z), (Z, Y)],
                        undirected=[(X, Z)],
                    ),
                )
            ],
        ),
        dict(
            id_in=[
                Identification.from_expression(
                    query=P(Y @ X),
                    estimand=P(X, Y),
                    graph=NxMixedGraph.from_edges(
                        directed=[(X, Y)],
                    ),
                )
            ],
            id_out=[
                Identification.from_expression(
                    query=P(Y @ X),
                    estimand=P(Y | X),
                    graph=NxMixedGraph.from_edges(
                        directed=[(X, Y)],
                    ),
                )
            ],
        ),
    ],
)

line_7_example = Example(
    name="ID Line 7 example, figure 5a and b",
    reference="Shpitser, I., & Pearl, J. (2008). Complete Identification Methods for the Causal Hierarchy.",
    graph=NxMixedGraph.from_edges(directed=[(X, Y1), (W1, Y1)], undirected=[(W1, Y1)]),
    identifications=[
        dict(
            id_in=[
                Identification.from_expression(
                    query=P(Y1 @ [X, W1]),
                    estimand=P(X, Y1, W1),
                    graph=NxMixedGraph.from_edges(
                        directed=[(X, Y1), (W1, X)], undirected=[(W1, Y1)]
                    ),
                )
            ],
            id_out=[
                Identification.from_expression(
                    query=P(Y1 @ W1),
                    estimand=P(Y1 | [X, W1]) * P(W1),
                    graph=NxMixedGraph.from_edges(undirected=[(W1, Y1)]),
                )
            ],
        )
    ],
)

figure_6a = Example(
    name="Causal graph with identifiable conditional effect P(y|do(x),z)",
    reference="Shpitser, I., & Pearl, J. (2008). Complete Identification Methods for the Causal Hierarchy.",
    graph=NxMixedGraph.from_edges(directed=[(X, Z), (Z, Y)], undirected=[(X, Z)]),
    identifications=[
        dict(
            id_in=[
                Identification.from_parts(
                    outcomes={Y},
                    treatments={X},
                    conditions={Z},
                    estimand=P(X, Y, Z),
                    graph=NxMixedGraph.from_edges(directed=[(X, Z), (Z, Y)], undirected=[(X, Z)]),
                ),
            ],
            id_out=[
                Identification.from_expression(
                    query=P(Y @ (X, Z)),
                    estimand=P(Y | (X, Z)) / Sum.safe(expression=P(Y | (X, Z)), ranges=(Y,)),
                    graph=NxMixedGraph.from_edges(directed=[(X, Z), (Z, Y)], undirected=list()),
                ),
            ],
        )
    ],
)

tikka_unidentifiable_graph = Example(
    name="Tikka's unidentifiable example",
    reference="Tikka, S. (2020). Identifying Counterfactual Queries with the R Package cfid",
    graph=NxMixedGraph.from_edges(
        directed=[(X, W), (W, Y), (D, Z), (Z, Y), (X, Y)], undirected=[(X, Y)]
    ),
)

tikka_unidentifiable_cfgraph = Example(
    name="Tikka's unidentifiable example",
    reference="Tikka, S. (2020). Identifying Counterfactual Queries with the R Package cfid",
    graph=NxMixedGraph.from_edges(
        directed=[(X @ -x, W @ -x), (W @ -x, Y @ -x), (D, Z), (Z, Y @ -x), (X, Y)],
        undirected=[(X, Y @ -x)],
    ),
)


figure_9a = Example(
    name="Original causal diagram",
    reference="Shpitser, I., & Pearl, J. (2008). Complete Identification Methods for the Causal Hierarchy.",
    graph=NxMixedGraph.from_edges(directed=[(X, W), (W, Y), (D, Z), (Z, Y)], undirected=[(X, Y)]),
)

figure_9b = Example(
    name="Parallel worlds graph for :math:`P(y_x|x', x_d, d)`",
    reference="Shpitser, I., & Pearl, J. (2008). Complete Identification Methods for the Causal Hierarchy.",
    graph=NxMixedGraph.from_edges(
        directed=[
            (X @ -X, W @ -X),
            (W @ -X, Y @ -X),
            (D @ -X, Z @ -X),
            (Z @ -X, Y @ -X),
            (X, W),
            (W, Y),
            (D, Z),
            (Z, Y),
            (X @ D, W @ D),
            (W @ D, Y @ D),
            (D @ D, Z @ D),
            (Z @ D, Y @ D),
        ],
        undirected=[
            (X, Y),
            (X @ D, X),
            (Y @ -X, Y),
            (Y, Y @ D),
            (Y @ D, Y @ -X),
            (X, Y @ -X),
            (X @ D, Y),
            (X, Y @ D),
            (X @ D, Y @ -X),
            (X @ D, Y @ D),
            (D @ -X, D),
            (W @ -X, W),
            (W, W @ D),
            (W @ D, W @ -X),
            (Z @ -X, Z),
            (Z, Z @ D),
            (Z @ -X, Z @ D),
        ],
    ),
)

figure_9c = Example(
    name="Counterfactual graph for :math:`P(y_x | x', z_d, d)`",
    reference="Shpitser, I., & Pearl, J. (2008). Complete Identification Methods for the Causal Hierarchy.",
    graph=NxMixedGraph.from_edges(
        directed=[(X @ -X, W @ -X), (W @ -X, Y @ -X), (D, Z), (Z, Y @ -X)],
        undirected=[(X, Y @ -X)],
    ),
)

tikka_figure_2 = Example(
    name=r"Figure 2: A graph for the example on identifiability of a conditional counterfacual "
    r"query :math:`P(y_x|z_x\wedge x')`",
    reference="Tikka, S (2022) Identifiying Counterfactual Queries with the R package cfid",
    graph=NxMixedGraph.from_edges(directed=[(X, Z), (X, Y), (Z, Y)], undirected=[(X, Z)]),
)

tikka_figure_5 = Example(
    name=r"Figure 5: Counterfactual graph :math:`G'` for :math:`y_x\wedge x'\wedge z_d\wedge d`",
    reference="Tikka, S (2022) Identifiying Counterfactual Queries with the R package cfid",
    graph=NxMixedGraph.from_edges(
        nodes=(X, Y @ -x, D, Z, X @ -x, W @ -x),
        directed=[(D, Z), (Z, Y @ -x), (X @ -x, W @ -x), (W @ -x, Y @ -x)],
        undirected=[(X, Y @ -x)],
    ),
)


tikka_figure_6a = Example(
    name=r"Figure 6a: Parallel worlds graph for :math:`y_x\wedge z_x\wedge x'` (the counterfactual graph)",
    reference="Tikka, S (2022) Identifiying Counterfactual Queries with the R package cfid",
    graph=NxMixedGraph.from_edges(
        directed=[(X, Z), (Z, Y), (X, Y), (X @ -x, Z @ -x), (Z @ -x, Y @ -x), (X @ -x, Y @ -x)],
        undirected=[(X, Z), (X, Z @ -x), (Z, Z @ -x), (Y, Y @ -x)],
    ),
)

tikka_figure_6b = Example(
    name=r"Figure 6b: Parallel worlds graph for :math:`y_{x,z}\wedge x'` (the counterfactual graph)",
    reference="Tikka, S (2022) Identifiying Counterfactual Queries with the R package cfid",
    graph=NxMixedGraph.from_edges(
        directed=[
            (X, Z),
            (Z, Y),
            (X, Y),
            (Z @ (-x, -z), Y @ (-x, -z)),
            (X @ (-x, -z), Y @ (-x, -z)),
        ],
        undirected=[(X, Z), (Y, Y @ (-x, -z))],
    ),
)

figure_9d = Example(
    name="Counterfactual graph resulting from application of make_counterfactual_graph() with"
    " joint distribution from which :math:`P(y_{x,z}|x')` is derived, namely  :math:`P(y_{x,z}, x')`",
    reference="Shpitser, I., & Pearl, J. (2008). Complete Identification Methods for the Causal Hierarchy.",
    graph=NxMixedGraph.from_edges(
        nodes=(X, X @ (-X, -Z), Z @ (-X, -Z), W @ (-X, -Z), Y @ (-X, -Z)),
        directed=[
            (X @ (-X, -Z), W @ (-X, -Z)),
            (Z @ (-X, -Z), Y @ (-X, -Z)),
            (W @ (-X, -Z), Y @ (-X, -Z)),
        ],
        undirected=[(X, Y @ (-X, -Z))],
    ),
)

figure_9e = Example(
    name="Counterfactual graph for :math:`P(Y @ (~X, Z) | X)`",
    reference="Shpitser, I., & Pearl, J. (2008). Complete Identification Methods for the Causal Hierarchy.",
    graph=NxMixedGraph.from_edges(
        nodes=(D, X, X @ (~X, Z), Z @ (~X, Z), W @ (~X, Z), Y @ (~X, Z)),
        directed=[(D, Z), (X @ (~X, Z), W @ (~X, Z)), (Z, Y @ (~X, Z)), (W @ (~X, Z), Y @ (~X, Z))],
        undirected=[(X, Y @ (~X, Z))],
    ),
)

figure_11a = Example(
    name="Intermediate graph obtained by **make-cg** in constructing the"
    " counterfactual graph for for :math:`P(y_x|x', z_d, d)` from Figure 9b",
    reference="Shpitser, I., & Pearl, J. (2008). Complete Identification Methods for the Causal Hierarchy.",
    graph=NxMixedGraph.from_edges(
        directed=[
            (X @ -X, W @ -X),
            (W @ -X, Y @ -X),
            (D, Z @ -X),
            (Z @ -X, Y @ -X),
            (X, W),
            (W, Y),
            (D, Z),
            (Z, Y),
            (X, W @ D),
            (W @ D, Y @ D),
            (D @ D, Z @ D),
            (Z @ D, Y @ D),
        ],
        undirected=[
            (X, Y),
            (Y @ -X, Y),
            (Y, Y @ D),
            (Y @ D, X),
            (X, Y @ -X),
            (Y @ D, Y @ -X),
            (W @ -X, W),
            (W, W @ D),
            (W @ D, W @ -X),
            (Z @ -X, Z),
            (Z, Z @ D),
            (Z @ -X, Z @ D),
        ],
    ),
)

figure_11b = Example(
    name="Intermediate graph obtained by **make-cg** in constructing the"
    " counterfactual graph for for :math:`P(y_x|x', z_d, d)` from Figure 9b",
    reference="Shpitser, I., & Pearl, J. (2008). Complete Identification Methods for the Causal Hierarchy.",
    graph=NxMixedGraph.from_edges(
        directed=[
            (X @ -X, W @ -X),
            (W @ -X, Y @ -X),
            (D, Z),
            (Z, Y @ -X),
            (Z, Y @ D),
            (Z, Y),
            (X, W),
            (W, Y),
            (W, Y @ D),
        ],
        undirected=[
            (X, Y),
            (Y @ -X, Y),
            (Y, Y @ D),
            (Y @ D, X),
            (Y @ D, Y @ -X),
            (X, Y @ -X),
            (X, Y @ D),
            (W @ -X, W),
        ],
    ),
)

figure_11c = Example(
    name="Intermediate graph obtained by **make-cg** in constructing the counterfactual"
    " graph for for :math:`P(y_x|x', z_d, d)` from Figure 9b",
    reference="Shpitser, I., & Pearl, J. (2008). Complete Identification Methods for the Causal Hierarchy.",
    graph=NxMixedGraph.from_edges(
        directed=[
            (X @ -X, W @ -X),
            (W @ -X, Y @ -X),
            (D, Z),
            (Z, Y @ -X),
            (Z, Y),
            (X, W),
            (W, Y),
        ],
        undirected=[
            (X, Y),
            (Y @ -X, Y),
            (X, Y @ -X),
            (W @ -X, W),
        ],
    ),
)

cyclic_directed_example = Example(
    name="Cyclic directed graph",
    reference="out of the mind of JZ and ZW",
    graph=NxMixedGraph.from_edges(directed=[(A, B), (A, C), (B, A)]),
)
#: Treatment: X
#: Outcome: Y
identifiability_1 = NxMixedGraph.from_edges(
    directed=[
        (Z1, Z2),
        (Z1, Z3),
        (Z2, X),
        (Z3, X),
        (Z4, X),
        (Z4, Z5),
        (Z3, Y),
        (X, Y),
        (Z3, Y),
    ],
)
identifiability_1_example = Example(
    name="Identifiability 1",
    reference='J. Pearl. 2009. "Causality: Models, Reasoning and Inference.'
    ' 2nd ed." Cambridge University Press, p. 80.',
    graph=identifiability_1,
    conditional_independencies=(
        DSeparationJudgement.create(X, Z1, [Z2, Z3]),
        DSeparationJudgement.create(X, Z5, [Z4]),
        DSeparationJudgement.create(Y, Z1, [X, Z3, Z4]),
        DSeparationJudgement.create(Y, Z2, [X, Z1, Z3]),
        DSeparationJudgement.create(Y, Z4, [X, Z3, Z5]),
        DSeparationJudgement.create(Z1, Z4),
        DSeparationJudgement.create(Z1, Z5),
        DSeparationJudgement.create(Z2, Z3, [Z1]),
        DSeparationJudgement.create(Z2, Z4),
        DSeparationJudgement.create(Z2, Z5),
        DSeparationJudgement.create(Z3, Z5),
        DSeparationJudgement.create(Y, Z5, [X, Z3]),
        DSeparationJudgement.create(Z3, Z4),
    ),
)

#: Treatment: X
#: Outcome: Y
identifiability_2 = NxMixedGraph.from_edges(
    directed=[
        (Z1, Z2),
        (Z1, Z3),
        (Z2, X),
        (Z3, X),
        (X, W0),
        (W0, Y),
        (Z4, Z3),
        (Z4, Z5),
        (Z5, Y),
        (X, W1),
        (W1, W2),
        (W2, Y),
        (Z4, Z3),
        (Z3, Y),
    ],
    undirected=[
        (Z1, X),
        (Z2, Z3),
        (Z3, Z5),
        (Z4, Y),
    ],
)

identifiability_2_example = Example(
    name="Identifiability 2",
    reference="E. Bareinboim modification of Identifiability 1.",
    graph=identifiability_2,
    verma_constraints=[
        VermaConstraint(
            rhs_cfactor=Q[Z5](Z4, Z5),
            rhs_expr=Sum[u_3, Z4](P(Z5 | (u_3, Z4)) * P(Z4) * P(u_3)),
            lhs_cfactor=Sum[Z3](Q[Z3, Z5](Z1, Z4, Z3, Z5)),
            lhs_expr=Sum[Z3](P(Z5 | (Z1, Z2, Z3, Z4)) * P(Z3 | (Z1, Z2, Z4))),
            variables=(Z1,),
        ),
        VermaConstraint(
            rhs_cfactor=Q[Z5](Z4, Z5),
            rhs_expr=Sum[u_3, Z4](P(Z5 | (u_3, Z4)) * P(Z4) * P(u_3)),
            lhs_cfactor=(Q[Z2, Z5](Z1, Z4, Z2, Z5) / Sum[Z5](Q[Z2, Z5](Z1, Z4, Z2, Z5))),
            lhs_expr=(
                Sum[Z3](P(Z5 | (Z1, Z2, Z3, Z4)) * P(Z3 | (Z1, Z4, Z2)) * P(Z2 | (Z1, Z4)))
                / Sum[Z3, Z5](P(Z5 | (Z1, Z4, Z2, Z3)) * P(Z3 | (Z1, Z4, Z2)) * P(Z2 | (Z1, Z4)))
            ),
            variables=(Z1, Z2),
        ),
    ],
    conditional_independencies=[
        DSeparationJudgement.create(W0, W1, [X]),
        DSeparationJudgement.create(W0, W2, [X]),
        DSeparationJudgement.create(W0, Z1, [X]),
        DSeparationJudgement.create(W0, Z2, [X]),
        DSeparationJudgement.create(W0, Z3, [X]),
        DSeparationJudgement.create(W0, Z4, [X]),
        DSeparationJudgement.create(W0, Z5, [X]),
        DSeparationJudgement.create(W1, Y, [W0, W2, Z3, Z4, Z5]),
        DSeparationJudgement.create(W1, Z1, [X]),
        DSeparationJudgement.create(W1, Z2, [X]),
        DSeparationJudgement.create(W1, Z3, [X]),
        DSeparationJudgement.create(W1, Z4, [X]),
        DSeparationJudgement.create(W1, Z5, [X]),
        DSeparationJudgement.create(W2, X, [W1]),
        DSeparationJudgement.create(W2, Z1, [W1]),
        DSeparationJudgement.create(W2, Z2, [W1]),
        DSeparationJudgement.create(W2, Z3, [W1]),
        DSeparationJudgement.create(W2, Z4, [W1]),
        DSeparationJudgement.create(W2, Z5, [W1]),
        DSeparationJudgement.create(X, Y, [W0, W2, Z3, Z4, Z5]),
        DSeparationJudgement.create(X, Z4, [Z1, Z2, Z3]),
        DSeparationJudgement.create(X, Z5, [Z1, Z2, Z3]),
        DSeparationJudgement.create(Y, Z1, [W0, W2, Z3, Z4, Z5]),
        DSeparationJudgement.create(Y, Z2, [W0, W2, Z3, Z4, Z5]),
        DSeparationJudgement.create(Z1, Z4),
        DSeparationJudgement.create(Z1, Z5),
        DSeparationJudgement.create(Z2, Z4),
        DSeparationJudgement.create(Z2, Z5),
    ],
)

#: The Identifiability 3 example
#: Treatment: X
#: Outcome: Y
#: Reference: J. Pearl. 2009. "Causality: Models, Reasoning and Inference. 2nd ed." Cambridge University Press, p. 92.
identifiability_3 = NxMixedGraph.from_edges(
    directed=[
        (Z2, X),
        (Z2, Z1),
        (Z2, Z3),
        (X, Z1),
        (Z3, Y),
        (Z1, Y),
    ],
    undirected=[
        (Z2, X),
        (Z2, Y),
        (X, Z3),
        (X, Y),
    ],
)

#: The Identifiability 4 example
#: Treatment: X
#: Outcome: Y
#: Reference: J. Pearl. 2009. "Causality: Models, Reasoning and Inference. 2nd ed." Cambridge University Press, p. 92.
identifiability_4 = NxMixedGraph.from_edges(
    directed=[
        (X, Z1),
        (X, Y),
        (Z1, Z2),
        (Z1, Y),
        (Z2, Y),
    ],
    undirected=[
        (X, Z2),
        (Z1, Y),
    ],
)

#: The Identifiability 5 example
#: Treatment: X1, X2
#: Outcome: Y
#: Reference: J. Pearl. 2009. "Causality: Models, Reasoning and Inference. 2nd ed." Cambridge University Press, p. 119.
identifiability_5 = NxMixedGraph.from_edges(
    directed=[
        ("X1", Z),
        ("X1", Y),
        ("X1", "X2"),
        (Z, "X2"),
        ("X2", Y),
    ],
    undirected=[
        ("X1", Z),
        (Z, Y),
    ],
)

#: The Identifiability 6 example
#: Treatment: X1, X2
#: Outcome: Y
#: Reference: J. Pearl. 2009. "Causality: Models, Reasoning and Inference. 2nd ed." Cambridge University Press, p. 125.
identifiability_6 = NxMixedGraph.from_edges(
    directed=[
        (Z1, "X1"),
        ("X1", "X2"),
        ("X2", Y),
        (Z2, Y),
    ],
    undirected=[
        (Z1, Z2),
        (Z1, "X2"),
        (Z2, "X2"),
    ],
)

#: The Identifiability 7 example
#: Treatment: X
#: Outcome: Y
#: Reference: J. Tian. 2002. "Studies in Causal Reasoning and Learning." p. 90.
identifiability_7 = NxMixedGraph.from_edges(
    directed=[
        (W1, W2),
        ("W3", "W4"),
        (W2, X),
        ("W4", X),
        (X, Y),
    ],
    undirected=[
        (W1, X),
        (W1, Y),
        (W1, "W3"),
        ("W3", W2),
        ("W3", "W5"),
        ("W5", "W4"),
    ],
)

# TODO Recoverability 1/2 - what is the S node?
# TODO Transportability 1/2 - what are the box nodes?
# TODO g-Identifiability examples
# TODO g-Transportability examples


#: The Verma 1 example
#: Treatment: V3
#: Outcome: V4
#: Reference: T. Verma and J. Pearl. 1990. "Equivalence and Synthesis of Causal Models." In P. Bonissone et al., eds.,
#: Proceedings of the 6th Conference on Uncertainty in Artificial Intelligence. Cambridge, MA: AUAI Press, p. 257.
verma_1 = NxMixedGraph.from_str_edges(
    directed=[
        ("V1", "V2"),
        ("V2", "V3"),
        ("V3", "V4"),
    ],
    undirected=[
        ("V2", "V4"),
    ],
)

#: The Verma 2 example
#: Treatment: V1
#: Outcome: V5
#: Reference: J. Tian. 2002. "Studies in Causal Reasoning and Learning." p. 70.
verma_2 = NxMixedGraph.from_str_edges(
    directed=[
        ("V1", "V2"),
        ("V2", "V3"),
        ("V3", "V4"),
        ("V4", "V5"),
    ],
    undirected=[
        ("V1", "V3"),
        ("V2", "V4"),
        ("V3", "V5"),
    ],
)

#: The Verma 3 example
#: Treatment: V1
#: Outcome: V5
#: Reference: J. Tian. 2002. "Studies in Causal Reasoning and Learning." p. 59.
verma_3 = NxMixedGraph.from_str_edges(
    directed=[
        ("V1", "V2"),
        ("V2", "V3"),
        ("V3", "V4"),
        ("V4", "V5"),
    ],
    undirected=[
        ("V1", "V5"),
        ("V1", "V3"),
        ("V2", "V4"),
    ],
)

#: The Verma 4 example
#: Treatment: V1
#: Outcome: V5
#: Reference: E. Bareinboim modification of Verma 2.
verma_4 = NxMixedGraph.from_str_edges(
    directed=[
        ("V1", "V2"),
        ("V2", "V3"),
        ("V3", "V4"),
        ("V4", "V5"),
    ],
    undirected=[
        ("V1", "V5"),
        ("V1", "V3"),
        ("V2", "V4"),
        ("V3", "V5"),
    ],
)

#: The Verma 5 example
#: Treatment: V1
#: Outcome: V5
#: Reference: E. Bareinboim modification of Verma 2.
verma_5 = NxMixedGraph.from_str_edges(
    directed=[
        ("V1", "V2"),
        ("V2", "V3"),
        ("V3", "V4"),
        ("V4", "V5"),
        ("V5", "V6"),
    ],
    undirected=[
        ("V0", "V1"),
        ("V0", "V6"),
        ("V1", "V5"),
        ("V1", "V3"),
        ("V2", "V4"),
    ],
)

#: The z-Identifiability 1 example
#: Treatment: X
#: Outcome: Y
#: Z*: Z
#: Reference: E. Bareinboim and J. Pearl. 2012. "Causal Inference by Surrogate Experiments: z-Identifiability." In
#: Nando de Freitas and K. Murphy., eds., Proceedings of the 28th Conference on Uncertainty in Artificial Intelligence.
#: Corvallis, OR: AUAI Press, p. 114.
z_identifiability_1 = NxMixedGraph.from_edges(
    directed=[
        (Z, X),
        (X, Y),
    ],
    undirected=[
        (Z, X),
        (Z, Y),
    ],
)

#: The z-Identifiability 2 example
#: Treatment: X
#: Outcome: Y
#: Z*: Z
#: Reference: E. Bareinboim and J. Pearl. 2012. "Causal Inference by Surrogate Experiments: z-Identifiability." In
#: Nando de Freitas and K. Murphy., eds., Proceedings of the 28th Conference on Uncertainty in Artificial Intelligence.
#: Corvallis, OR: AUAI Press, p. 114.
z_identifiability_2 = NxMixedGraph.from_edges(
    directed=[
        (Z, X),
        (X, Y),
    ],
    undirected=[
        (X, Y),
        (Z, Y),
    ],
)

#: The z-Identifiability 3 example
#: Treatment: X
#: Outcome: Y
#: Z*: Z
#: Reference: E. Bareinboim and J. Pearl. 2012. "Causal Inference by Surrogate Experiments: z-Identifiability." In
#: Nando de Freitas and K. Murphy., eds., Proceedings of the 28th Conference on Uncertainty in Artificial Intelligence.
#: Corvallis, OR: AUAI Press, p. 114.
z_identifiability_3 = NxMixedGraph.from_edges(
    directed=[
        (Z, Y),
        (X, Y),
    ],
    undirected=[
        (X, Z),
        (Z, Y),
    ],
)

#: The Identifiability (Linear) 1 example
#: Treatment: X
#: Outcome: Y
#: Reference: J. Pearl. 2009. "Causality: Models, Reasoning and Inference. 2nd ed." Cambridge University Press, p. 153.
identifiability_linear_1 = NxMixedGraph.from_edges(
    directed=[
        (X, Z),
        (X, W),
        (W, Y),
        (Z, Y),
    ],
    undirected=[
        (X, Z),
        (W, Y),
    ],
)

d_separation_example = Example(
    name="D-separation example",
    reference="http://web.mit.edu/jmn/www/6.034/d-separation.pdf",
    graph=NxMixedGraph.from_edges(
        directed=[
            (AA, C),
            (B, C),
            (C, D),
            (C, E),
            (D, F),
            (F, G),
        ],
    ),
    conditional_independencies=[
        DSeparationJudgement.create(AA, B),
        DSeparationJudgement.create(AA, D, [C]),
        DSeparationJudgement.create(AA, E, [C]),
        DSeparationJudgement.create(AA, F, [C]),
        DSeparationJudgement.create(AA, G, [C]),
        DSeparationJudgement.create(B, D, [C]),
        DSeparationJudgement.create(B, E, [C]),
        DSeparationJudgement.create(B, F, [C]),
        DSeparationJudgement.create(B, G, [C]),
        DSeparationJudgement.create(C, F, [D]),
        DSeparationJudgement.create(C, G, [D]),
        DSeparationJudgement.create(D, E, [C]),
        DSeparationJudgement.create(D, G, [F]),
        DSeparationJudgement.create(E, F, [C]),
        DSeparationJudgement.create(E, G, [C]),
    ],
)

asia_example = Example(
    name="Asia dataset",
    reference="https://www.bnlearn.com/documentation/man/asia.html",
    graph=NxMixedGraph.from_edges(
        directed=[
            (Variable(u), Variable(v))
            for u, v in [
                ("Asia", "Tub"),
                ("Smoke", "Lung"),
                ("Smoke", "Bronc"),
                ("Tub", "Either"),
                ("Lung", "Either"),
                ("Either", "Xray"),
                ("Either", "Dysp"),
                ("Bronc", "Dysp"),
            ]
        ],
    ),
    data=pd.read_csv(ASIA_PATH).replace({"yes": 1, "no": -1}),
)

figure_2a_example = Example(
    name="Shpitser et al. (2008), Figure 2A",
    reference="Shpitser, I., & Pearl, J. (2008). Complete Identification Methods for the Causal Hierarchy. "
    "Journal of Machine Learning Research.",
    graph=NxMixedGraph.from_edges(
        directed=[(X, Y)],
    ),
)

figure_2b_example = Example(
    name="Shpitser et al. (2008), Figure 2B",
    reference="Shpitser, I., & Pearl, J. (2008). Complete Identification Methods for the Causal Hierarchy. "
    "Journal of Machine Learning Research.",
    graph=NxMixedGraph.from_edges(
        directed=[(X, Y), (X, Z), (Z, Y)],
        undirected=[(Y, Z)],
    ),
)

complete_hierarchy_figure_2c_example = Example(
    name="Shpitser et al (2008) figure 2d",
    reference="Shpitser, I., & Pearl, J. (2008). Complete Identification Methods for the Causal Hierarchy. "
    "Journal of Machine Learning Research.",
    graph=NxMixedGraph.from_edges(
        directed=[
            (X, Y),
            (Z, X),
            (Z, Y),
        ],
        undirected=[(X, Z)],
    ),
)

complete_hierarchy_figure_2d_example = Example(
    name="Shpitser et al (2008) figure 2d",
    reference="Shpitser, I., & Pearl, J. (2008). Complete Identification Methods for the Causal Hierarchy. "
    "Journal of Machine Learning Research.",
    graph=NxMixedGraph.from_edges(
        directed=[
            (X, Y),
            (Z, X),
            (Z, Y),
        ],
        undirected=[(X, Z)],
    ),
)

complete_hierarchy_figure_2e_example = Example(
    name="Shpitser et al (2008) figure 2e",
    reference="Shpitser, I., & Pearl, J. (2008). Complete Identification Methods for the Causal Hierarchy. "
    "Journal of Machine Learning Research.",
    graph=NxMixedGraph.from_edges(
        directed=[
            (X, Z),
            (Z, Y),
        ],
        undirected=[(X, Y)],
    ),
)

complete_hierarchy_figure_3a_example = Example(
    name="Shpitser et al 2008 figure 3a",
    reference="Shpitser, I., & Pearl, J. (2008). Complete Identification Methods for the Causal Hierarchy."
    " Journal of Machine Learning Research.",
    graph=NxMixedGraph.from_edges(
        directed=[(X, Y1), (W1, X), (W2, Y2)],
        undirected=[(W1, W2), (W1, Y1), (W1, Y2), (X, W2)],
    ),
)

id_sir_example = Example(
    name="Identifiable SIR",
    reference="ASKEM",
    graph=NxMixedGraph.from_str_edges(
        directed=[
            ("Infected", "Hospitalized"),
            ("Hospitalized", "Died"),
        ],
        undirected=[("Infected", "Died")],
    ),
)

nonid_sir_example = Example(
    name="Non-Identifiable SIR",
    reference="ASKEM",
    graph=NxMixedGraph.from_str_edges(
        directed=[
            ("Infected", "Died"),
        ],
        undirected=[("Infected", "Died")],
    ),
)

igf_example = Example(
    name="IGF Graph",
    reference="Jeremy Zucker, Sara Mohammad-Taheri, Kaushal Paneri, Somya Bhargava, Pallavi Kolambkar"
    ", Craig Bakker, Jeremy Teuton, Charles Tapley Hoyt, Kristie Oxford, Robert Ness, and Olga Vitek. 2021."
    "Leveraging Structured Biological Knowledge for Counterfactual Inference: a Case Study of Viral Pathogenesis"
    "- IEEE Journals & Magazine. IEEE Transactions on Big Data (January 2021).",
    graph=NxMixedGraph.from_str_edges(
        nodes=["SOS", "Ras", "Raf", "AKT", "Mek", "Erk", "PI3K"],
        directed=[
            ("SOS", "Ras"),
            ("Ras", "PI3K"),
            ("Ras", "Raf"),
            ("PI3K", "AKT"),
            ("AKT", "Raf"),
            ("Raf", "Mek"),
            ("Mek", "Erk"),
        ],
        undirected=[("SOS", "PI3K")],
    ),
    example_queries=[Query.from_str(treatments="SOS", outcomes="Erk")],
)

sars_large_example = Example(
    name="SARS-CoV-2 Graph",
    reference="Jeremy Zucker, Sara Mohammad-Taheri, Kaushal Paneri, Somya Bhargava, Pallavi Kolambkar"
    ", Craig Bakker, Jeremy Teuton, Charles Tapley Hoyt, Kristie Oxford, Robert Ness, and Olga Vitek. 2021."
    "Leveraging Structured Biological Knowledge for Counterfactual Inference: a Case Study of Viral Pathogenesis"
    "- IEEE Journals & Magazine. IEEE Transactions on Big Data (January 2021).",
    graph=NxMixedGraph.from_str_edges(
        nodes=[
            "SARS_COV2",
            "ACE2",
            "Ang",
            "AGTR1",
            "ADAM17",
            "Toci",
            "Sil6r",
            "EGF",
            "TNF",
            "EGFR",
            "PRR",
            "NFKB",
            "IL6STAT3",
            "IL6AMP",
            "cytok",
            "Gefi",
        ],
        directed=[
            ("SARS_COV2", "ACE2"),
            ("ACE2", "Ang"),
            ("Ang", "AGTR1"),
            ("AGTR1", "ADAM17"),
            ("ADAM17", "EGF"),
            ("ADAM17", "TNF"),
            ("ADAM17", "Sil6r"),
            ("SARS_COV2", "PRR"),
            ("PRR", "NFKB"),
            ("EGFR", "NFKB"),
            ("TNF", "NFKB"),
            ("Sil6r", "IL6STAT3"),
            ("Toci", "Sil6r"),
            ("NFKB", "IL6AMP"),
            ("IL6AMP", "cytok"),
            ("IL6STAT3", "IL6AMP"),
            ("EGF", "EGFR"),
            ("Gefi", "EGFR"),
        ],
        undirected=[
            ("SARS_COV2", "Ang"),
            ("ADAM17", "Sil6r"),
            ("PRR", "NFKB"),
            ("EGF", "EGFR"),
            ("EGFR", "TNF"),
            ("EGFR", "IL6STAT3"),
        ],
    ),
    example_queries=[
        Query.from_str(treatments="Sil6r", outcomes="cytok"),
        Query.from_str(treatments="EGFR", outcomes="cytok"),
    ],
)

SARS_SMALL_GRAPH = NxMixedGraph.from_str_edges(
    directed=[
        ("ADAM17", "EGFR"),
        ("ADAM17", "TNF"),
        ("ADAM17", "Sil6r"),
        ("EGFR", "cytok"),
        ("TNF", "cytok"),
        ("Sil6r", "IL6STAT3"),
        ("IL6STAT3", "cytok"),
    ],
    undirected=[
        ("ADAM17", "cytok"),
        ("ADAM17", "Sil6r"),
        ("EGFR", "TNF"),
        ("EGFR", "IL6STAT3"),
    ],
)

sars_small_example = Example(
    name="SARS-CoV-2 Small Graph",
    reference="Sara!",  # FIXME
    graph=SARS_SMALL_GRAPH,
    generate_data=generate_data_for_covid_case_study,
    example_queries=[Query.from_str(outcomes="cytok", treatments="EGFR")],
)

tikka_trso_figure_8_graph = NxMixedGraph.from_edges(
    undirected=[(X1, Y1), (Z, W), (Z, X2)],
    directed=[
        (X1, Y1),
        (X1, Y2),
        (W, Y1),
        (W, Y2),
        (Z, Y1),
        (Z, X2),
        (X2, Y2),
        (Z, Y2),
    ],
)
tikka_trso_figure_8 = Example(
    name="Tikka TRSO Figure 8",
    reference="https://arxiv.org/abs/1806.07172",
    graph=tikka_trso_figure_8_graph,
)


cancer_example = Example(
    name="Smoking and Cancer",
    reference="https://github.com/y0-causal-inference/y0/pull/183",
    graph=NxMixedGraph.from_edges(directed=[(S, T), (T, C), (S, C)], undirected=[(S, T)]),
    generate_data=generate_data_for_smoke_cancer,
    example_queries=[Query.from_str(outcomes="C", treatments="S")],
)


examples = [v for name, v in locals().items() if name.endswith("_example")]



# -*- coding: utf-8 -*-

"""Test the probability DSL."""

import unittest
from typing import Optional

from y0.dsl import (
    A,
    B,
    C,
    CounterfactualVariable,
    D,
    Distribution,
    Element,
    Intervention,
    One,
    P,
    Product,
    Q,
    R,
    S,
    Sum,
    T,
    Variable,
    W,
    X,
    Y,
    Z,
    Zero,
)
from y0.parser import parse_y0

V = Variable("V")


class TestDSL(unittest.TestCase):
    """Tests for the stringifying instances of the probability DSL."""

    def assert_exp(self, expression: Element, s: Optional[str] = None):
        """Test an element can be parsed, serialized, then again."""
        e = expression.to_y0()
        if s:
            self.assertEqual(s, e, msg=repr(e))
        reconstituted = parse_y0(e)
        self.assertEqual(
            expression,
            reconstituted,
            msg=f"\nError in reconstitution.\nExpected: {expression.to_y0()}\nActual:   {reconstituted.to_y0()}",
        )

    def assert_text(self, s: str, expression: Element):
        """Assert the expression when it is converted to a string."""
        self.assertIsInstance(s, str)
        self.assertIsInstance(hash(expression), int)  # can the expression be hashed?
        self.assertIsInstance(expression.to_text(), str)
        self.assertIsInstance(expression.to_latex(), str)
        self.assertIsInstance(expression._repr_latex_(), str)
        self.assertEqual(s, expression.to_text(), msg=f"Expression: {repr(expression)}")
        if not isinstance(expression, (Distribution, Intervention)):
            self.assert_exp(expression)

    def test_variable(self):
        """Test the variable DSL object."""
        self.assert_text("A", Variable("A"))
        self.assert_text("A", A)  # shorthand for testing purposes

    def test_stop_the_madness(self):
        """Test that a variable can not be named "P"."""
        with self.assertRaises(ValueError):
            _ = Variable("P")
        with self.assertRaises(ValueError):
            _ = Variable("Q")

    def test_intervention(self):
        """Test the invervention DSL object."""
        self.assert_text("W*", Intervention("W", star=True))
        self.assert_text("W", Intervention("W", star=False))
        self.assert_text("W", W)  # shorthand for testing purposes
        self.assert_text("W", -W)  # An intervention from variable W

        # inversions using the unary ~ operator
        self.assert_text("W", ~Intervention("W", star=True))
        self.assert_text("W*", ~Intervention("W", star=False))
        self.assert_text("W*", ~W)

    def test_counterfactual_variable(self):
        """Test the Counterfactual Variable DSL object."""
        # Normal instantiation
        self.assert_text("Y_{W}", CounterfactualVariable("Y", interventions=(-W,)))
        self.assert_text("Y_{W*}", CounterfactualVariable("Y", interventions=(~W,)))

        # Instantiation with list-based operand to matmul @ operator
        self.assert_text("Y_{W}", Variable("Y") @ [W])
        self.assert_text("Y_{W}", Y @ [W])
        self.assert_text("Y_{W*}", Variable("Y") @ [~W])
        self.assert_text("Y_{W*}", Y @ [~W])

        # Instantiation with two variables
        self.assert_text(
            "Y_{W*, X}",
            CounterfactualVariable(
                "Y", interventions=(~Intervention("W", star=False), Intervention("X", star=False))
            ),
        )

        # Instantiation with matmul @ operator and single operand
        self.assert_text("Y_{W}", Y @ Intervention("W", star=False))
        self.assert_text("Y_{W*}", Y @ ~Intervention("W", star=False))

        # Instantiation with matmul @ operator and list operand
        self.assert_text("Y_{W*, X}", Y @ [X, ~W])

        # Instantiation with matmul @ operator (chained)
        self.assert_text("Y_{W*, X}", Y @ X @ ~W)

    def test_star_counterfactual(self):
        """Tests for generalized counterfactual variables."""
        for expr, expected in [
            # Single variable
            (P(Y @ X), "P[X](Y)"),
            (P(Y @ -X), "P[X](Y)"),
            (P(Y @ ~X), "P[+X](Y)"),
            (P(Y @ +X), "P[+X](Y)"),
            #
            (P(-Y @ X), "P[X](-Y)"),
            (P(-Y @ -X), "P[X](-Y)"),
            (P(-Y @ ~X), "P[+X](-Y)"),
            (P(-Y @ +X), "P[+X](-Y)"),
            #
            (P(~Y @ X), "P[X](+Y)"),
            (P(~Y @ -X), "P[X](+Y)"),
            (P(~Y @ ~X), "P[+X](+Y)"),
            (P(~Y @ +X), "P[+X](+Y)"),
            #
            (P(+Y @ X), "P[X](+Y)"),
            (P(+Y @ -X), "P[X](+Y)"),
            (P(+Y @ ~X), "P[+X](+Y)"),
            (P(+Y @ +X), "P[+X](+Y)"),
            # Interventions can live inside the conditions
            (P(Y @ X | ~X, ~Y), "P(Y @ -X | +X, +Y)"),
            (P(Y @ -X | ~X, ~Y), "P(Y @ -X | +X, +Y)"),
            (P(Y @ +X | ~X, ~Y), "P(Y @ +X | +X, +Y)"),
            (P(Y @ ~X | ~X, ~Y), "P(Y @ +X | +X, +Y)"),
            #
            (P(~(Y @ ~X) | X, Y), "P(+Y @ +X | X, Y)"),
            (P(~Y @ ~X | X, Y), "P(+Y @ +X | X, Y)"),  # should be same as above
            # Counterfactuals can live inside the conditions
            (P(Y @ X | ~X, Z @ D, D), "P(Y @ -X | D, +X, Z @ -D)"),
            (P(Y @ X | ~X, Z @ D, -D), "P(Y @ -X | -D, +X, Z @ -D)"),
            (P(Y @ X | ~X, Z @ D, +D), "P(Y @ -X | +D, +X, Z @ -D)"),
            (P(Y @ X | ~X, Z @ D, ~D), "P(Y @ -X | +D, +X, Z @ -D)"),
        ]:
            with self.subTest(expr=expected):
                self.assert_exp(expr, expected)

    def test_counterfactual_errors(self):
        """Test that if two variables with the same name are given, an error is raised, regardless of star state."""
        Y @ Intervention("X", star=True) @ Intervention("X", star=True)
        Y @ Intervention("X", star=False) @ Intervention("X", star=False)
        with self.assertRaises(ValueError):
            Y @ Intervention("X", star=True) @ Intervention("X", star=False)
        with self.assertRaises(ValueError):
            Y @ Intervention("X", star=False) @ Intervention("X", star=True)

    def test_conditional_distribution(self):
        """Test the :class:`Distribution` DSL object."""
        # Normal instantiation
        self.assert_text("A | B", Distribution((A,), (B,)))

        # Instantiation with list-based operand to or | operator
        self.assert_text("A | B", Variable("A") | (B,))
        self.assert_text("A | B", A | (B,))

        # # Instantiation with two variables
        self.assert_text("A | B, C", A | [B, C])

        # Instantiation with or | operator and single operand
        self.assert_text("A | B", Variable("A") | B)
        self.assert_text("A | B", A | B)

        # Instantiation with or | operator (chained)
        self.assert_text("A | B, C", A | B | C)

        # Counterfactual uses work basically the same.
        #  Note: @ binds more tightly than |, but it's probably better to use parentheses
        self.assert_text("Y_{W} | B", (Y @ W) | B)
        self.assert_text("Y_{W} | B", Y @ W | B)
        self.assert_text("Y_{W} | B, C", Y @ W | B | C)
        self.assert_text("Y_{W, X*} | B, C", Y @ W @ ~X | B | C)
        self.assert_text("Y_{W, X*} | B_{N*}, C", Y @ W @ ~X | B @ Intervention("N", star=True) | C)

    def test_joint_distribution(self):
        """Test the JointProbability DSL object."""
        self.assert_text("A, B", Distribution((A, B)))
        self.assert_text("A, B", A & B)
        self.assert_text("A, B, C", Distribution((A, B, C)))
        self.assert_text("A, B, C", A & B & C)

    def test_probability(self):
        """Test generation of probabilities."""
        # Make sure there are children
        with self.assertRaises(ValueError):
            Distribution(tuple())

        self.assert_text("P(A)", P(A))
        self.assert_text("P(A)", P("A"))
        self.assert_text("P(A)", P(Distribution((A,))))

        # Test markov kernel with single parent (AKA has only one child variable)
        self.assert_text("P(A | B)", P(A | B))
        self.assert_text("P(A | B)", P(Distribution((A,), (B,))))
        self.assert_text("P(A | B)", P(A | [B]))

        # Test markov kernel with multiple parents
        self.assert_text("P(A | B, C)", P(A | B, C))
        self.assert_text("P(A | B, C)", P(Distribution((A,), (B,)) | C))
        self.assert_text("P(A | B, C)", P(A | [B, C]))
        self.assert_text("P(A | B, C)", P(A | B | C))
        self.assert_text("P(A | B, C)", P(A | B & C))

        # Test simple joint distributions
        self.assert_text("P(A, B)", P((A, B)))
        self.assert_text("P(A, B)", P([A, B]))
        self.assert_text("P(A, B)", P({A, B}))
        self.assert_text("P(A, B)", P(A, B))
        self.assert_text("P(A, B)", P(A & B))
        self.assert_text("P(A, B)", P("A", "B"))
        self.assert_text("P(A, B)", P(["A", "B"]))
        self.assert_text("P(A, B, C)", P(A & B & C))
        self.assert_text("P(A, B, C)", P("A", "B", "C"))
        self.assert_text("P(A, B, C)", P(["A", "B", "C"]))
        self.assert_text("P(A, B, C)", P((name for name in "ABC")))
        self.assert_text("P(A, B, C)", P(name for name in "ABC"))
        self.assert_text("P(A, B, C)", P((Variable(name) for name in "ABC")))
        self.assert_text("P(A, B, C)", P(Variable(name) for name in "ABC"))

        # Test mixed with single conditional
        self.assert_text("P(A, B | C)", P(A, B | C))
        self.assert_text("P(A, B | C)", P(Distribution((A, B), (C,))))
        self.assert_text("P(A, B | C)", P(Distribution((A, B), (C,))))
        self.assert_text("P(A, B | C)", P(Distribution((A, B)) | C))
        self.assert_text("P(A, B | C)", P(A & B | C))

        # Test mixed with multiple conditionals
        self.assert_text("P(A, B | C, D)", P(A, B | C, D))
        self.assert_text("P(A, B | C, D)", P(Distribution((A, B), (C, D))))
        self.assert_text("P(A, B | C, D)", P(Distribution((A, B)) | C | D))
        self.assert_text("P(A, B | C, D)", P(Distribution((A, B), (C,)) | D))
        self.assert_text("P(A, B | C, D)", P(A & B | C | D))
        self.assert_text("P(A, B | C, D)", P(A & B | (C, D)))
        self.assert_text("P(A, B | C, D)", P(A & B | Distribution((C, D))))
        self.assert_text("P(A, B | C, D)", P(A & B | C & D))

    def test_conditioning_errors(self):
        """Test erroring on conditionals."""
        for expression in [
            Distribution((B,), (C,)),
            Distribution((B, C), (D,)),
            Distribution((B, C), (D, W)),
        ]:
            with self.assertRaises(TypeError):
                _ = A | expression
            with self.assertRaises(TypeError):
                _ = X & Y | expression

    def test_sum(self):
        """Test the Sum DSL object."""
        with self.assertRaises(TypeError):
            Sum(P(A), (~B,))
        with self.assertRaises(TypeError):
            Sum(P(A), (B @ C,))
        with self.assertRaises(TypeError):
            Sum(P(A), tuple())
        with self.assertRaises(ValueError):
            Sum(P(A), frozenset())

        # Sum with one variable
        self.assert_text(
            "[ sum_{S} P(A | B) P(C | D) ]",
            Sum[S](P(A | B) * P(C | D)),
        )
        # Sum with two variables
        self.assert_text(
            "[ sum_{S, T} P(A | B) P(C | D) ]",
            Sum[S, T](P(A | B) * P(C | D)),
        )

        # CRAZY sum syntax! pycharm doesn't like this usage of __class_getitem__ though so idk if we'll keep this
        self.assert_text(
            "[ sum_{S} P(A | B) P(C | D) ]",
            Sum[S](P(A | B) * P(C | D)),
        )
        self.assert_text(
            "[ sum_{S, T} P(A | B) P(C | D) ]",
            Sum[S, T](P(A | B) * P(C | D)),
        )

        # Sum with sum inside
        self.assert_text(
            "[ sum_{S, T} P(A | B) [ sum_{R} P(C | D) ] ]",
            Sum[S, T](P(A | B) * Sum[R](P(C | D))),
        )

    def test_q(self):
        """Test the Q DSL object."""
        self.assert_text("Q[A](X)", Q[A](X))  # type: ignore
        self.assert_text("Q[A, B](X)", Q[A, B](X))  # type: ignore
        self.assert_text("Q[A](X, Y)", Q[A](X, Y))  # type: ignore
        self.assert_text("Q[A, B](X, Y)", Q[A, B](X, Y))  # type: ignore

    def test_jeremy(self):
        """Test assorted complicated objects from Jeremy."""
        self.assert_text(
            "[ sum_{W} P(D) P(W_{X*}) P(X, Y_{W, Z*}) P(Z_{D}) ]",
            Sum[W](P(X, (Y @ ~Z @ W)) * P(D) * P(Z @ D) * P(W @ ~X)),
        )

        self.assert_text(
            "[ sum_{W} P(W_{X*}) P(X, Y_{W, Z*}) ]",
            Sum[W](P(X, Y @ ~Z @ W) * P(W @ ~X)),
        )

        self.assert_text(
            "[ sum_{D} P(D) P(W_{X*}) P(X, Y_{W, Z*}) P(Z_{D}) ]",
            Sum[D](P(X, Y @ ~Z @ W) * P(D) * P(Z @ D) * P(W @ ~X)),
        )

    def test_api(self):
        """Test the high-level API for getting variables."""
        for expression, variables in [
            (A @ B, {A @ B, -B}),
            (A @ ~B, {A @ ~B, ~B}),
            (P(A @ B), {A @ B, -B}),
            (P(A @ ~B), {A @ ~B, ~B}),
            (P(A @ ~B) * P(C), {A @ ~B, ~B, C}),
            (Sum[S](P(A @ ~B) * P(C)), {A @ ~B, ~B, C, S}),
            (Sum[S, T](P(A @ ~B) * P(C)), {A @ ~B, ~B, C, S, T}),
            (One() / P(A @ B), {A @ B, -B}),
            (P(B) / P(A @ ~B), {A @ ~B, B, ~B}),
            (P(Y | X) * P(X) / P(Y), {X, Y}),
            (Q[A, B](C, D), {A, B, C, D}),  # type: ignore
        ]:
            with self.subTest(expression=str(expression)):
                self.assertEqual(variables, expression.get_variables())


class TestCounterfactual(unittest.TestCase):
    """Tests for counterfactuals."""

    def test_event_semantics(self):
        """Check DSL semantics."""
        for expr, counterfactual_star, intervention_star in [
            (X @ X, None, False),
            (-X @ X, False, False),
            (+X @ X, True, False),
            (~X @ X, True, False),
        ]:
            with self.subTest(expr=expr.to_y0()):
                self.assertIsInstance(expr, CounterfactualVariable)
                self.assertEqual(counterfactual_star, expr.star)
                self.assertEqual(1, len(expr.interventions))
                self.assertEqual(intervention_star, list(expr.interventions)[0].star)

    def test_event_failures(self):
        """Check for failure to determine tautology/inconsistent."""
        for expr in [
            # Opposite variable
            X @ Y,
            X @ +Y,
            X @ -Y,
            X @ ~Y,
            # Same variable
            X @ X,
            X @ +X,
            X @ -X,
            X @ ~X,
        ]:
            with self.subTest(expr=expr.to_y0()):
                self.assertIsInstance(expr, CounterfactualVariable)
                self.assertFalse(expr.is_event())
                with self.assertRaises(ValueError):
                    expr.has_tautology()
                with self.assertRaises(ValueError):
                    expr.is_inconsistent()

    def test_tautology(self):
        """Check for tautologies."""
        for expr, status in [
            # Different Variable
            (~X @ Y, False),
            (~X @ ~Y, False),
            # Same variable, self.star is False
            (-X @ X, True),
            (-X @ -X, True),
            (-X @ +X, False),
            (-X @ ~X, False),
            # Same variable, self.star is True
            (+X @ X, False),
            (+X @ -X, False),
            (+X @ +X, True),
            (+X @ ~X, True),
            # Same variable, self.star is True
            (~X @ X, False),
            (~X @ -X, False),
            (~X @ +X, True),
            (~X @ ~X, True),
        ]:
            with self.subTest(expr=expr.to_y0()):
                self.assertIsInstance(expr, CounterfactualVariable)
                self.assertTrue(expr.is_event())
                self.assertEqual(status, expr.has_tautology())

    def test_inconsistent(self):
        """Check for tautologies."""
        for expr, status in [
            # Different Variable
            (~X @ Y, False),
            (~X @ ~Y, False),
            # Same variable, self.star is False
            (-X @ X, False),
            (-X @ -X, False),
            (-X @ +X, True),
            (-X @ ~X, True),
            # Same variable, self.star is True
            (+X @ X, True),
            (+X @ -X, True),
            (+X @ +X, False),
            (+X @ ~X, False),
            # Same variable, self.star is True
            (~X @ X, True),
            (~X @ -X, True),
            (~X @ +X, False),
            (~X @ ~X, False),
        ]:
            with self.subTest(expr=expr.to_y0()):
                self.assertIsInstance(expr, CounterfactualVariable)
                self.assertTrue(expr.is_event())
                self.assertEqual(status, expr.is_inconsistent())

    def test_counterfactual_y0(self):
        """Test compressed output."""
        self.assertEqual("P[X](Y)", P(Y @ X).to_y0())
        self.assertEqual("P[X](Y)", P[X](Y).to_y0())
        self.assertEqual("P[X](Y)", P(Y @ -X).to_y0())
        self.assertEqual("P[X](Y)", P[-X](Y).to_y0())
        self.assertEqual("P[+X](Y)", P(Y @ ~X).to_y0())
        self.assertEqual("P[+X](Y)", P[~X](Y).to_y0())

        self.assertEqual("P[+X](Y)", P(Y @ +X).to_y0())
        self.assertEqual("P[+X](Y)", P[+X](Y).to_y0())

        # Two variables, same intervention
        self.assertEqual("P[X](Y, Z)", P(Y @ X, Z @ X).to_y0())
        self.assertEqual("P[X](Y, Z)", P[X](Y, Z).to_y0())
        self.assertEqual("P[X](Y, Z)", P(Y @ -X, Z @ -X).to_y0())
        self.assertEqual("P[X](Y, Z)", P[-X](Y, Z).to_y0())
        self.assertEqual("P[+X](Y, Z)", P(Y @ ~X, Z @ ~X).to_y0())
        self.assertEqual("P[+X](Y, Z)", P[~X](Y, Z).to_y0())

        # Two variables, mixed intervention
        self.assertEqual("P(Y @ -X, Z @ -A)", P(Y @ X, Z @ A).to_y0())
        self.assertEqual("P(Y @ -X, Z @ +Z)", P(Y @ -X, Z @ +Z).to_y0())


class TestSafeConstructors(unittest.TestCase):
    """Test that the .safe() constructors work properly."""

    def test_do2_intervention(self):
        """Test the do-calculus level two interventions."""
        self.assertEqual(P[X](Y), P(Y @ X))
        self.assertEqual(P[X](Y, Z), P(Y @ X & Z @ X))
        self.assertEqual(P[X](Y | Z), P(Y @ X | Z @ X))

        # stack interventions mixed with $L_3$ notation
        self.assertEqual(P[X](Y @ Z), P(Y @ Z @ X))

        # mixed with $L_3$, where each variable can have different kinds of interventions
        self.assertEqual(P[X](Y @ Z | W), P(Y @ Z @ X | W @ X))

    def test_sum(self):
        """Test the :meth:`Sum.safe` constructor."""
        self.assertEqual(Sum(P(X, Y), frozenset([X])), Sum.safe(P(X, Y), (X,)))
        self.assertEqual(Sum(P(X, Y), frozenset([X])), Sum.safe(P(X, Y), [X]))
        self.assertEqual(Sum(P(X, Y), frozenset([X])), Sum.safe(P(X, Y), {X}))
        self.assertEqual(Sum(P(X, Y), frozenset([X])), Sum.safe(P(X, Y), X))

        self.assertEqual(
            Sum(P(X, Y, Z), frozenset([X, Y])), Sum.safe(P(X, Y, Z), (v for v in [X, Y]))
        )

    def test_product(self):
        """Test the :meth:`Product.safe` constructor."""
        p1 = P(X, Y)
        p2 = P(Z)
        self.assertEqual(p1, Product.safe(p1))
        self.assertEqual(p1, Product.safe([p1]))
        self.assertEqual(Product((p1, p2)), Product.safe((p1, p2)))
        self.assertEqual(Product((p1, p2)), Product.safe([p1, p2]))
        self.assertEqual(Product((P(X), P(Y))), Product.safe(P(v) for v in [X, Y]))

        self.assertEqual(One(), Product.safe([]))
        self.assertEqual(One(), Product.safe([One()]))
        self.assertEqual(One(), Product.safe([One(), One()]))

        self.assertEqual(Product((P(X), P(Y))), Product.safe((One(), P(X), P(Y))))
        self.assertEqual(Product((P(X), P(Y))), Product.safe((P(X), P(Y))))
        self.assertEqual(Product((P(X), P(Y))), Product.safe((P(X), One(), P(Y))))
        self.assertEqual(Product((P(X), P(Y))), Product.safe((P(X), P(Y), One(), One())))
        self.assertEqual(Product((P(X), P(Y))), Product.safe((P(X), One(), P(Y), One(), One())))

        self.assertEqual(P(X), Product.safe((One(), P(X))))
        self.assertEqual(P(X), Product.safe((P(X),)))
        self.assertEqual(P(X), Product.safe((P(X), One())))
        self.assertEqual(P(X), Product.safe((P(X), One(), One())))
        self.assertEqual(P(X), Product.safe((P(X), One(), One(), One())))

        self.assertEqual(Zero(), Product.safe((P(X), Zero())))
        self.assertEqual(Zero(), Product.safe((P(X), Zero(), One())))
        self.assertEqual(Zero(), Product.safe((Zero(), One())))


zero = Zero()


class TestZero(unittest.TestCase):
    """Tests for zero."""

    exprs = [
        One(),
        Zero(),
        P(A),
        P(A) * P(B),
        P(A) / P(B),
        Sum.safe(P(A), [A]),
    ]

    def test_divide_failure(self):
        """Test failure is thron on division by zero."""
        for expr in self.exprs:
            with self.subTest(expr=expr), self.assertRaises(ZeroDivisionError):
                expr / zero

    def test_identity(self):
        """Test that zero divided by anything is zero."""
        for expr in self.exprs:
            if isinstance(expr, Zero):
                continue  # would raise divides by zero
            with self.subTest(expr=expr):
                self.assertEqual(zero, zero / expr)

    def test_multiply(self):
        """Test other operations."""
        for expr in self.exprs:
            with self.subTest(expr=expr.to_y0(), direction="right"):
                self.assertEqual(zero, zero * expr, msg=f"Got {zero * expr}")
            with self.subTest(expr=expr.to_y0(), direction="left"):
                self.assertEqual(zero, expr * zero, msg=f"Got {expr * zero}")



# -*- coding: utf-8 -*-

"""Data structures."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Iterable, NamedTuple, Optional, Tuple, Union

import pandas as pd

from .dsl import Expression, Variable

__all__ = [
    "VermaConstraint",
    "DSeparationJudgement",
]


class VermaConstraint(NamedTuple):
    """Represent a Verma constraint."""

    lhs_cfactor: Expression
    lhs_expr: Expression
    rhs_cfactor: Expression
    rhs_expr: Expression
    variables: Tuple[Variable, ...]

    @classmethod
    def from_element(cls, element) -> VermaConstraint:
        """Extract content from each element in the vector returned by `verma.constraint`.

        :param element: An element in the in the vector returned by `verma.constraint`
        :returns: A Verma constraint tuple for the given element

        .. seealso:: Extracting from R objects https://rpy2.github.io/doc/v3.4.x/html/vector.html#extracting-items
        """
        from .parser import parse_causaleffect
        from .r_utils import _extract, _parse_vars

        return cls(
            rhs_cfactor=parse_causaleffect(_extract(element, "rhs.cfactor")),
            rhs_expr=parse_causaleffect(_extract(element, "rhs.expr")),
            lhs_cfactor=parse_causaleffect(_extract(element, "lhs.cfactor")),
            lhs_expr=parse_causaleffect(_extract(element, "lhs.expr")),
            variables=_parse_vars(element),
        )


@dataclass(frozen=True)
class DSeparationJudgement:
    """
    Record if a left/right pair are d-separated given the conditions.

    By default, acts like a boolean, but also caries evidence graph.
    """

    separated: bool
    left: Variable
    right: Variable
    conditions: Tuple[Variable, ...]

    @classmethod
    def create(
        cls,
        left: Variable,
        right: Variable,
        conditions: Optional[Iterable[Variable]] = None,
        *,
        separated: bool = True,
    ) -> DSeparationJudgement:
        """Create a d-separation judgement in canonical form."""
        left, right = sorted([left, right], key=str)
        if conditions is None:
            conditions = tuple()
        conditions = tuple(sorted(set(conditions), key=str))
        return cls(separated, left, right, conditions)

    def __bool__(self) -> bool:
        return self.separated

    @property
    def is_canonical(self) -> bool:
        """Return if the conditional independency is in canonical form."""
        return (
            self.left < self.right
            and isinstance(self.conditions, tuple)
            and tuple(sorted(self.conditions, key=str)) == (self.conditions)
        )

    def test(
        self, df: pd.DataFrame, boolean: bool = False, method: Optional[str] = None, **kwargs
    ) -> Union[Tuple[float, int, float], bool]:
        """Test for conditional independence, given some data.

        :param df: A dataframe.
        :param boolean: Should results be returned as a pre-cutoff boolean?
        :param method: Conditional independence from :mod:`pgmpy` to use. If none,
            defaults to :func:`pgmpy.estimators.CITests.cressie_read`.
        :param kwargs: Additional kwargs to pass to the estimator function
        :returns:
            Tests the null hypothesis that X is independent of Y given Zs.
            If ``boolean=False``, returns a three-tuple of chi, dof, p_value.
            If ``boolean=True``, make sure you also set ``significance_level=0.05`` or your preferred
            value, then returns simply a boolean if the test fails.
        :raises ValueError: if any parts of the judgement aren't in the dataframe's
            columns
        :raises ImportError: If pgmpy is not available. Install with pip.
        """
        try:
            from pgmpy.estimators import CITests
        except ImportError as e:
            raise ImportError("Calculating falsifications requires `pip install pgmpy`.") from e

        tests = {
            "pearson": CITests.pearsonr,
            "chi-square": CITests.chi_square,
            "cressie_read": CITests.cressie_read,
            "freeman_tuckey": CITests.freeman_tuckey,
            "g_sq": CITests.g_sq,
            "log_likelihood": CITests.log_likelihood,
            "modified_log_likelihood": CITests.modified_log_likelihood,
            "power_divergence": CITests.power_divergence,
            "neyman": CITests.neyman,
        }

        if self.left.name not in df.columns:
            raise ValueError(
                f"left variable {self.left.name} ({type(self.left.name)}) not in columns {list(df.columns)}"
            )
        if self.right.name not in df.columns:
            raise ValueError(
                f"right variable {self.right.name} ({type(self.right.name)}) not in columns {df.columns}"
            )
        for c in self.conditions:
            if c.name not in df.columns:
                raise ValueError(
                    f"conditional {c.name} ({type(c.name)}) not in columns {df.columns}"
                )

        func = tests[method or "cressie_read"]
        return func(
            X=self.left.name,
            Y=self.right.name,
            Z={condition.name for condition in self.conditions},
            data=df,
            boolean=boolean,
            **kwargs,
        )



"""Examples for SARS-CoV-2 and COVID19."""

import numpy as np
import pandas as pd

from y0.dsl import Variable

__all__ = [
    "generate_data_for_covid_case_study",
]


def _r_exp(x):
    return 1 / (1 + np.exp(x))


def generate_data_for_covid_case_study(
    num_samples: int, treatments: dict[Variable, float] | None = None, *, seed: int | None = None
) -> pd.DataFrame:
    """Generate testing data for the SARS-CoV-2 small graph.

    :param num_samples: The number of samples to generate. Try 1000.
    :param treatments: An optional dictionary of the values to fix each variable to.
    :param seed: An optional random seed for reproducibility purposes
    :returns: A pandas Dataframe with columns corresponding
        to the variable names SARS-CoV-2 small graph
    """
    if treatments is None:
        treatments = {}
    generator = np.random.default_rng(seed)

    u_adam17_sil6r_value = generator.normal(loc=40.0, scale=10.0, size=num_samples)
    u_il6_stat_egfr_value = generator.normal(loc=44.0, scale=10.0, size=num_samples)
    u_tnf_egfr_value = generator.normal(loc=40.0, scale=10.0, size=num_samples)
    u_adam17_cytok_value = generator.normal(loc=44.0, scale=10.0, size=num_samples)

    beta0_adam17 = -1
    beta_u_adam17_cytok = 0.04
    beta_u_adam17_sil6r = 0.04
    adam17 = generator.normal(
        loc=100
        * _r_exp(
            -beta0_adam17
            - u_adam17_cytok_value * beta_u_adam17_cytok
            - u_adam17_sil6r_value * beta_u_adam17_sil6r
        ),
        scale=1,
        size=num_samples,
    )

    beta0_sil6r = -1.9
    beta_adam17_to_sil6r = 0.03
    beta_u_adam17_sil6r = 0.05
    sil6r_value = generator.normal(
        loc=100
        * _r_exp(
            -beta0_sil6r
            - adam17 * beta_adam17_to_sil6r
            - u_adam17_sil6r_value * beta_u_adam17_sil6r
        ),
        scale=1,
    )

    beta0_tnf = -1.8
    beta_adam17_to_tnf = 0.05
    beta_u_tnf_egfr = 0.06
    tnf = generator.normal(
        loc=100
        * _r_exp(-beta0_tnf - adam17 * beta_adam17_to_tnf - u_tnf_egfr_value * beta_u_tnf_egfr),
        scale=1,
    )

    beta0_egfr = -1.9
    beta_adam17_egfr = 0.03
    beta_u_il6_stat_egfr = -0.04
    beta_u_tnf_egfr = 0.02
    if Variable("EGFR") in treatments:
        egfr = np.full(num_samples, treatments[Variable("EGFR")])
    else:
        p = _r_exp(
            -beta0_egfr
            - adam17 * beta_adam17_egfr
            - u_il6_stat_egfr_value * beta_u_il6_stat_egfr
            - u_tnf_egfr_value * beta_u_tnf_egfr
        )
        egfr = generator.binomial(1, p, size=num_samples)

    beta0_il6_stat3 = -1.6
    beta_u_il6_stat_egfr = -0.05
    beta_sil6r_to_il6_stat3 = 0.04
    il6_stat3 = generator.normal(
        loc=100
        * _r_exp(
            -beta0_il6_stat3
            - u_il6_stat_egfr_value * beta_u_il6_stat_egfr
            - sil6r_value * beta_sil6r_to_il6_stat3
        ),
        scale=1,
    )

    beta0_cytok = -1.9
    beta_il6_stat3_tocytok = 0.02
    beta_egfr_tocytok = 0.06
    beta_tnf_tocytok = 0.01
    beta_u_adam17_cytok = 0.01
    cytok = generator.normal(
        loc=100
        * _r_exp(
            -beta0_cytok
            - il6_stat3 * beta_il6_stat3_tocytok
            - egfr * beta_egfr_tocytok
            - tnf * beta_tnf_tocytok
            - u_adam17_cytok_value * beta_u_adam17_cytok
        ),
        scale=1,
    )

    data = {
        "ADAM17": adam17,
        "Sil6r": sil6r_value,
        "TNF": tnf,
        "EGFR": egfr,
        "IL6STAT3": il6_stat3,
        "cytok": cytok,
    }
    df = pd.DataFrame(data)
    return df



# -*- coding: utf-8 -*-

"""Tests for the identify algorithm."""

import itertools as itt
import unittest

import y0.examples
from y0.algorithm.identify import Identification, Query, Unidentifiable, idc, identify
from y0.algorithm.identify.id_std import (
    line_1,
    line_2,
    line_3,
    line_4,
    line_5,
    line_6,
    line_7,
)
from y0.dsl import (
    W1,
    W2,
    Y1,
    Y2,
    Expression,
    M,
    P,
    Probability,
    Product,
    Sum,
    X,
    Y,
    Z,
    get_outcomes_and_treatments,
)
from y0.examples import (
    figure_6a,
    line_1_example,
    line_2_example,
    line_3_example,
    line_4_example,
    line_5_example,
    line_6_example,
    line_7_example,
)
from y0.graph import NxMixedGraph
from y0.mutate import canonicalize

P_XY = P(X, Y)
P_XYZ = P(X, Y, Z)


class TestIdentify(unittest.TestCase):
    """Test cases from https://github.com/COVID-19-Causal-Reasoning/Y0/blob/master/ID_whittemore.ipynb."""

    def assert_identify(self, expected: Expression, graph: NxMixedGraph, query: Probability):
        """Assert the ID algorithm returns the expected result."""
        id_in = Identification(Query.from_expression(query), graph)
        self.assert_expr_equal(expected, identify(id_in))

    def assert_expr_equal(self, expected: Expression, actual: Expression) -> None:
        """Assert that two expressions are the same."""
        expected_outcomes, expected_treatments = get_outcomes_and_treatments(query=expected)
        actual_outcomes, actual_treatments = get_outcomes_and_treatments(query=actual)
        self.assertEqual(expected_treatments, actual_treatments)
        self.assertEqual(expected_outcomes, actual_outcomes)
        ordering = tuple(expected.get_variables())
        expected_canonical = canonicalize(expected, ordering)
        actual_canonical = canonicalize(actual, ordering)
        self.assertEqual(
            expected_canonical,
            actual_canonical,
            msg=f"\nExpected: {str(expected_canonical)}\nActual:   {str(actual_canonical)}",
        )

    def test_idc(self):
        r"""Test that the IDC algorithm works correctly."""
        for identification in figure_6a.identifications:
            id_in = identification["id_in"][0]
            id_out = identification["id_out"][0]
            self.assert_expr_equal(
                expected=id_out.estimand,
                actual=idc(id_in),
            )

    def test_line_1(self):
        r"""Test that line 1 of ID algorithm works correctly.

        If no action has been taken, the effect on :math:`\mathbf Y` is just the marginal of
        the observational distribution :math:`P(\mathbf v)` on :math:`\mathbf Y`.
        """
        for identification in line_1_example.identifications:
            self.assert_expr_equal(
                expected=identification["id_out"][0].estimand,
                actual=line_1(identification["id_in"][0]),
            )
            self.assert_expr_equal(
                identification["id_out"][0].estimand,
                identify(identification["id_in"][0]),
            )

    def test_line_2(self):
        r"""Test line 2 of the identification algorithm.

        2. If we are interested in the effect on :math:`\mathbf Y`, it is sufficient to restrict our
        attention on the parts of the model ancestral to :math:`\mathbf Y`.
        """
        for identification in line_2_example.identifications:
            id_out = identification["id_out"][0]
            self.assertEqual(
                id_out,
                line_2(identification["id_in"][0]),
            )
            self.assert_expr_equal(
                Sum.safe(expression=Sum.safe(expression=P(Y, X, Z), ranges=[X]), ranges=[Z]),
                identify(identification["id_in"][0]),
            )
        with self.assertRaises(ValueError):
            line_2(
                Identification.from_parts(
                    outcomes={Y},
                    treatments={X},
                    estimand=P(X, Y, Z),
                    graph=NxMixedGraph.from_edges(directed=[(X, Z), (Z, Y)]),
                )
            )

    def test_line_3(self):
        r"""Test line 3 of the identification algorithm.

        3. Forces an action on any node where such an action would have no effect on :math:`\mathbf Y`—assuming
        we already acted on :math:`\mathbf X`. Since actions remove incoming arrows, we can view line 3 as
        simplifying the causal graph we consider by removing certain arcs from the graph, without
        affecting the overall answer.
        """
        for identification in line_3_example.identifications:
            self.assertEqual(
                identification["id_out"][0],
                line_3(identification["id_in"][0]),
            )
            self.assert_expr_equal(
                P(Y | (X, Z)),
                identify(identification["id_in"][0]),
            )
        with self.assertRaises(ValueError):
            line_3(
                Identification.from_parts(
                    outcomes={Y},
                    treatments={X},
                    estimand=P(X, Y, Z),
                    graph=NxMixedGraph.from_edges(directed=[(X, Z), (Z, Y)]),
                )
            )

    def test_line_4(self):
        r"""Test line 4 of the identification algorithm.

        4. The key line of the algorithm, it decomposes the problem into a set of smaller problems
        using the key property of *c-component factorization* of causal models. If the entire graph
        is a single C-component already, further problem decomposition is impossible, and we must
        provide base cases. :math:`\mathbf{ID}` has three base cases.
        """
        for identification in line_4_example.identifications:
            actuals = line_4(identification["id_in"][0])
            expecteds = identification["id_out"]
            self.assertEqual(len(expecteds), len(actuals))
            match = []
            for expected, actual in itt.product(expecteds, actuals):
                if expected == actual:
                    match.append((expected, actual))
            self.assertEqual(len(expecteds), len(match), msg="Could not match identifications")

            # TODO @cthoyt
            self.assert_expr_equal(
                Sum.safe(
                    ranges=[M, Z],
                    expression=Product.safe(
                        [
                            P(M | (Z, X)),
                            P(Y | (M, Z, X)),
                            Sum.safe(expression=P(Z, X, M, Y), ranges=[X, M, Y]),
                        ]
                    ),
                ),
                identify(identification["id_in"][0]),
            )

        with self.assertRaises(ValueError):
            line_4(
                Identification.from_parts(
                    outcomes={Y},
                    treatments={X},
                    estimand=P(X, Y, Z),
                    graph=NxMixedGraph.from_edges(undirected=[(X, Y), (Y, Z)]),
                )
            )

    def test_line_5(self):
        r"""Test line 5 of the identification algorithm.

        5. Fails because it finds two C-components, the graph :math:`G` itself, and a subgraph :math:`S` that
        does not contain any :math:`\mathbf X` nodes. But that is exactly one of the properties of C-forests
        that make up a hedge. In fact, it turns out that it is always possible to recover a hedge
        from these two c-components.
        """
        for identification in line_5_example.identifications:
            with self.assertRaises(Unidentifiable):
                line_5(identification["id_in"][0])
            with self.assertRaises(Unidentifiable):
                identify(identification["id_in"][0])
        self.assertIsNone(
            line_5(
                Identification(
                    Query.from_expression(P(X, Y, Z)),
                    NxMixedGraph.from_edges(directed=[(X, Y), (Y, Z)]),
                )
            )
        )

    def test_line_6(self):
        r"""Test line 6 of the identification algorithm.

        6. Asserts that if there are no bidirected arcs from X to the other nodes in the current
        subproblem under consideration, then we can replace acting on X by conditioning, and thus
        solve the subproblem.
        """
        for identification in line_6_example.identifications:
            id_out = identification["id_out"][0]

            self.assert_expr_equal(
                expected=id_out.estimand,
                actual=line_6(identification["id_in"][0]),
            )
        with self.assertRaises(ValueError):
            line_6(line_7_example.identifications[0]["id_in"][0])

    def test_line_7(self):
        r"""Test line 2 of the identification algorithm.

        7. The most complex case where :math:`\mathbf X` is partitioned into two sets, :math:`\mathbf W` which
        contain bidirected arcs into other nodes in the subproblem, and :math:`\mathbf Z` which do not.
        In this situation, identifying :math:`P(\mathbf y|do(\mathbf x))` from :math:`P(v)` is equivalent to
        identifying :math:`P(\mathbf y|do(\mathbf w))` from :math:`P(\mathbf V|do(\mathbf z))`, since
        :math:`P(\mathbf y|do(\mathbf x)) = P(\mathbf y|do(\mathbf w), do(\mathbf z))`. But the term
        :math:`P(\mathbf V|do(\mathbf z))` is identifiable using the previous base case, so we can consider
        the subproblem of identifying :math:`P(\mathbf y|do(\mathbf w))`.
        """
        for identification in line_7_example.identifications:
            id_out = identification["id_out"][0]
            id_in = identification["id_in"][0]
            self.assertEqual(id_out, line_7(id_in))
            self.assert_expr_equal(
                Sum.safe(expression=P(Y1 | (W1, X)) * P(W1), ranges=[W1]), identify(id_in)
            )

    def test_figure_2a(self):
        """Test Figure 2A. from Shpitser *et al.*, (2008)."""
        graph = y0.examples.figure_2a_example.graph
        # expr = "[ sum_{} P(Y|X) ]"
        # frac_expr = P_XY / Sum[Y](P_XY)
        cond_expr = P(Y | X)
        self.assert_identify(cond_expr, graph, P(Y @ X))

    def test_figure_2b(self):
        """Test Figure 2B. from Shpitser *et al.*, (2008)."""
        graph = y0.examples.figure_2b_example.graph
        # expr = "[ sum_{Z} P(Z|X) P(Y|X,Z) ]"
        # frac_expr = Sum[Z](Sum[Y](P_XY) / (Sum[Z](Sum[Y](P_XY))) * (P_XY / Sum[Y](P_XY)))
        cond_expr = Sum[Z](P(Z | X) * P(Y | X, Z))
        self.assert_identify(cond_expr, graph, P(Y @ X))

    def test_figure_2c(self):
        """Test Figure 2C from Shpitser *et al.*, (2008)."""
        graph = y0.examples.complete_hierarchy_figure_2c_example.graph
        # expr = "[ sum_{Z} P(Z) P(Y|X,Z) ]"
        # frac_expr = Sum[Z](Sum[X, Y](P_XYZ) / (Sum[Z](Sum[X, Y](P_XYZ))) * (P_XYZ / Sum[Y](P_XYZ)))
        cond_expr = Sum[Z](P(Y | X, Z) * Sum[X, Y](P(X, Y, Z)))
        self.assert_identify(cond_expr, graph, P(Y @ X))

    def test_figure_2d(self):
        """Test Figure 2D from Shpitser *et al.*, (2008).

        .. note:: frac_expr = Sum[Z](Sum[X, Y](P_XYZ) * P_XYZ / Sum[Y](P_XYZ))
        """
        graph = y0.examples.complete_hierarchy_figure_2d_example.graph
        expr = Sum[Z](P(Y | X, Z) * Sum[X, Y](P(X, Y, Z)))
        self.assert_identify(expr, graph, P(Y @ X))

    def test_figure_2e(self):
        """Test Figure 2E from Shpitser *et al.*, (2008)."""
        graph = y0.examples.complete_hierarchy_figure_2e_example.graph
        # expr = "[ sum_{Z} [ sum_{} P(Z|X) ] [ sum_{} [ sum_{X} P(X) P(Y|X,Z) ] ] ]"
        # frac_expr = Sum[Z](Sum[Y](P_XYZ) / Sum[Z](Sum[Y](P_XYZ))) * Sum[X](
        #     P_XYZ * Sum[Y, Z](P_XYZ) / Sum[Y](P_XYZ) / Sum[X](Sum[Y, Z](P_XYZ))
        # )
        cond_expr = Sum[Z](Sum[X](P(Y | X, Z) * P(X)) * P(Z | X))
        self.assert_identify(cond_expr, graph, P(Y @ X))

    def test_figure_3a(self):
        """Test Figure 3A (A graph hedge-less for ``P(y1,y2|do(x))``) from Shpitser *et al.*, (2008)."""
        graph = y0.examples.complete_hierarchy_figure_3a_example.graph
        cond_expr = Sum[W2](
            Sum[W1, X, Y1, Y2](P(W1, W2, X, Y1, Y2)) * Sum[W1](P(W1) * P(Y1 | W1, X)) * P(Y2 | W2)
        )
        self.assert_identify(cond_expr, graph, P(Y1 @ X, Y2 @ X))



# -*- coding: utf-8 -*-

"""Implement of surrogate outcomes and transportability from https://arxiv.org/abs/1806.07172.

.. todo::

    high level documentation

    1. What problem are we trying to solve here?
    2. What's the difference between surrogate outcomes nad transportability?
    3. Real world example
"""

import logging
from copy import deepcopy
from dataclasses import dataclass
from typing import Collection, Dict, FrozenSet, Iterable, Optional, Set, Union, cast

from y0.algorithm.conditional_independencies import are_d_separated
from y0.dsl import (
    TARGET_DOMAIN,
    CounterfactualVariable,
    Distribution,
    Expression,
    Fraction,
    Intervention,
    One,
    Population,
    PopulationProbability,
    Probability,
    Product,
    Sum,
    Variable,
    Zero,
    _upgrade_variables,
)
from y0.graph import NxMixedGraph
from y0.mutate.canonicalize_expr import canonicalize

__all__ = [
    "identify_target_outcomes",
    "trso",
    "TransportQuery",
]

logger = logging.getLogger(__name__)


def get_nodes_to_transport(
    *,
    surrogate_interventions: Union[Set[Variable], Variable],
    surrogate_outcomes: Union[Set[Variable], Variable],
    graph: NxMixedGraph,
) -> Set[Variable]:
    """Identify which nodes the transport nodes should point to.

    :param surrogate_interventions: The interventions performed in an experiment.
    :param surrogate_outcomes: The outcomes observed in an experiment.
    :param graph: The graph of the target domain.
    :returns: A set of variables representing target domain nodes where transportability nodes should be added.
    """
    surrogate_interventions = set(_upgrade_variables(surrogate_interventions))
    surrogate_outcomes = set(_upgrade_variables(surrogate_outcomes))

    # Find the c_component with surrogate_outcomes
    c_component_surrogate_outcomes: Set[Variable] = set()
    for component in graph.districts():
        # Check if surrogate_outcomes is present in the current set
        if surrogate_outcomes.intersection(component):
            c_component_surrogate_outcomes.update(component)

    ancestors_surrogate_outcomes = graph.get_intervened_ancestors(
        surrogate_interventions, surrogate_outcomes
    )

    # Descendants of interventions in graph
    descendants_interventions = graph.descendants_inclusive(surrogate_interventions)

    return (descendants_interventions - surrogate_outcomes).union(
        c_component_surrogate_outcomes - ancestors_surrogate_outcomes
    )


_TRANSPORT_PREFIX = "T_"


def transport_variable(variable: Variable) -> Variable:
    """Create a transport Variable by adding the transport prefix to a variable.

    :param variable: variable that the transport node will point to
    :returns: Variable with _TRANSPORT_PREFIX and variable name
    :raises TypeError: If a non-standard variable is passed
    """
    if isinstance(variable, (CounterfactualVariable, Intervention)):
        raise TypeError
    return Variable(_TRANSPORT_PREFIX + variable.name)


def is_transport_node(node: Variable) -> bool:
    """Check if a Variable is a transport node.

    :param node: A node to evaluate.
    :returns: boolean True if node is a transport node, False otherwise.
    """
    return not isinstance(node, (CounterfactualVariable, Intervention)) and node.name.startswith(
        _TRANSPORT_PREFIX
    )


def get_transport_nodes(graph: NxMixedGraph) -> Set[Variable]:
    """Find all the transport nodes in a graph.

    :param graph: an NxMixedGraph which may have transport nodes
    :returns: Set containing all transport nodes in the graph
    """
    return {node for node in graph.nodes() if is_transport_node(node)}


def get_regular_nodes(graph: NxMixedGraph) -> Set[Variable]:
    """Find all the nodes in a graph which are not transport nodes.

    :param graph: an NxMixedGraph
    :returns: Set containing all nodes which are not transport nodes
    """
    return {node for node in graph.nodes() if not is_transport_node(node)}


def _c14n_safe(expression: Expression | None) -> Expression | None:
    if expression is None:
        return None
    return canonicalize(expression)


def create_transport_diagram(
    *,
    nodes_to_transport: Iterable[Variable],
    graph: NxMixedGraph,
) -> NxMixedGraph:
    """Create a NxMixedGraph identical to graph but with transport nodes added.

    :param nodes_to_transport: nodes which have transport nodes pointing to them.
    :param graph: The graph of the target domain.
    :returns: graph with transport nodes added
    """
    rv = NxMixedGraph()
    for node in graph.nodes():
        rv.add_node(node)
    for u, v in graph.directed.edges():
        rv.add_directed_edge(u, v)
    for u, v in graph.undirected.edges():
        rv.add_undirected_edge(u, v)
    for node in nodes_to_transport:
        transport_node = transport_variable(node)
        rv.add_directed_edge(transport_node, node)
    return rv


@dataclass
class TransportQuery:
    """A query used as output for surrogate_to_transport."""

    target_interventions: Set[Variable]
    target_outcomes: Set[Variable]
    graphs: Dict[Population, NxMixedGraph]
    domains: Set[Population]
    surrogate_interventions: Dict[Population, Set[Variable]]
    target_experiments: Set[Variable]


@dataclass
class TRSOQuery:
    """A query used for TRSO input."""

    target_interventions: Set[Variable]
    target_outcomes: Set[Variable]
    expression: Expression
    active_interventions: Set[Variable]
    domain: Population
    domains: Set[Population]
    graphs: Dict[Population, NxMixedGraph]
    surrogate_interventions: Dict[Population, Set[Variable]]


def surrogate_to_transport(
    *,
    graph: NxMixedGraph,
    target_outcomes: Set[Variable],
    target_interventions: Set[Variable],
    surrogate_outcomes: Dict[Population, Set[Variable]],
    surrogate_interventions: Dict[Population, Set[Variable]],
) -> TransportQuery:
    """Create transportability diagrams and query from a surrogate outcome problem.

    :param target_outcomes: A set of target variables for causal effects.
    :param target_interventions: A set of interventions for the target domain.
    :param graph: The graph of the target domain.
    :param surrogate_outcomes: A dictionary of outcomes in other populations
    :param surrogate_interventions: A dictionary of interventions in other populations
    :returns: An octuple representing the query transformation of a surrogate outcome query.
    :raises ValueError: if surrogate outcomes' and surrogate interventions' keys do not correspond
    """
    if set(surrogate_outcomes) != set(surrogate_interventions):
        raise ValueError("Inconsistent surrogate outcome and intervention domains")

    graphs = {
        domain: create_transport_diagram(
            graph=graph,
            nodes_to_transport=get_nodes_to_transport(
                surrogate_interventions=surrogate_interventions[domain],
                surrogate_outcomes=domain_outcomes,
                graph=graph,
            ),
        )
        for domain, domain_outcomes in surrogate_outcomes.items()
    }
    graphs[TARGET_DOMAIN] = graph

    return TransportQuery(
        target_interventions=target_interventions,
        target_outcomes=target_outcomes,
        graphs=graphs,
        domains=set(surrogate_outcomes),
        surrogate_interventions=surrogate_interventions,
        target_experiments=set(),
    )


def trso_line1(
    target_outcomes: Set[Variable],
    expression: Expression,
    graph: NxMixedGraph,
) -> Expression:
    """Return the probability in the case where no interventions are present.

    :param target_outcomes: A set of nodes that comprise our target outcomes.
    :param expression: The distribution in the current domain.
    :param graph: The graph with transport nodes in this domain.
    :returns: Sum over the probabilities of nodes other than target outcomes.
    """
    return Sum.safe(expression, get_regular_nodes(graph) - target_outcomes)


def trso_line2(
    query: TRSOQuery,
    outcomes_ancestors: Set[Variable],
) -> TRSOQuery:
    """Restrict the interventions and diagram to only include ancestors of target variables.

    :param query: A TRSO query
    :param outcomes_ancestors: the ancestors of target variables in transportability_diagram
    :returns: A TRSO query with modified attributes.
    """
    new_query = deepcopy(query)
    new_query.target_interventions.intersection_update(outcomes_ancestors)

    for domain, graph in query.graphs.items():
        outcome_ancestors_domain = graph.ancestors_inclusive(query.target_outcomes)
        new_query.graphs[domain] = graph.subgraph(outcome_ancestors_domain)

    new_query.expression = Sum.safe(
        query.expression,
        get_regular_nodes(query.graphs[query.domain]) - outcomes_ancestors,
        simplify=True,
    )
    if isinstance(new_query.expression, Probability):
        assert isinstance(new_query.expression, PopulationProbability)
        # it might be the case that these two are not the same, but
        # other parts of the algorithm clean it up. This isn't so
        # satisfying. Sorry!
        # if new_query.expression.population != new_query.domain:
        #     pass
        new_query.expression = PopulationProbability(
            population=new_query.domain,
            distribution=Distribution(
                children=new_query.expression.children,
            ),
        )
    return new_query


def trso_line3(query: TRSOQuery, additional_interventions: Set[Variable]) -> TRSOQuery:
    """Add nodes that will affect the outcome to the interventions of the query.

    :param query: A TRSO query
    :param additional_interventions: interventions to be added to target_interventions
    :returns: A TRSO query with modified attributes.
    """
    new_query = deepcopy(query)
    new_query.target_interventions.update(additional_interventions)
    return new_query


def trso_line4(
    query: TRSOQuery,
    components: Iterable[FrozenSet[Variable]],
) -> Dict[FrozenSet[Variable], TRSOQuery]:
    """Find the trso inputs for each C-component.

    :param query: A TRSO query
    :param components: Set of c_components of transportability_diagram without target_interventions
    :returns: Dictionary with components as keys TRSOQuery objects as values
    """
    graph = query.graphs[query.domain]
    rv = {}
    for component in components:
        new_query = deepcopy(query)
        new_query.target_outcomes = set(component)
        new_query.target_interventions = get_regular_nodes(graph) - component
        rv[component] = new_query
    return rv


def trso_line6(query: TRSOQuery) -> Dict[Population, TRSOQuery]:
    """Find the active interventions for each domain, remove available experiments from interventions.

    :param query: A TRSO query
    :returns: Dictionary with domains as keys TRSOQuery objects as values
    """
    expressions = {}
    for domain, graph in query.graphs.items():
        if domain == TARGET_DOMAIN:
            continue
        new_query = _line_6_helper(query, domain, graph)
        if new_query is not None:
            expressions[domain] = new_query
    return expressions


def _line_6_helper(
    query: TRSOQuery, domain: Population, graph: NxMixedGraph
) -> Optional[TRSOQuery]:
    """Perform d-separation check and then modify query active interventions.

    :param query: A TRSO query
    :param domain: A given population
    :param graph: A NxMixedGraph
    :returns: A TRSO query or None
    """
    surrogate_interventions = query.surrogate_interventions[domain]
    surrogate_intersect_target = surrogate_interventions.intersection(query.target_interventions)
    if not surrogate_intersect_target:
        return None
    if not all_transports_d_separated(
        graph,
        target_interventions=query.target_interventions,
        target_outcomes=query.target_outcomes,
    ):
        return None

    new_query = deepcopy(query)
    new_query.target_interventions = query.target_interventions - surrogate_interventions
    new_query.domain = domain
    new_query.graphs[new_query.domain] = graph.remove_nodes_from(surrogate_intersect_target)
    new_query.active_interventions = surrogate_intersect_target
    return new_query


def activate_domain_and_interventions(
    expression: Expression, interventions: Set[Variable], domain: Population
) -> Expression:
    """Intervene on the target variables of expression using the active interventions.

    :param expression: A probability expression.
    :param interventions: Set of active interventions
    :param domain: A given population
    :returns: A new expression, intervened
    :raises NotImplementedError: If an expression type that is not handled gets passed
    """
    if isinstance(expression, Probability):
        assert isinstance(expression, PopulationProbability)
        return PopulationProbability(
            population=domain,
            distribution=Distribution.safe(set(expression.children) - interventions),
        ).intervene(interventions)
    if isinstance(expression, Sum):
        # TODO need full integration test to trso() function that covers this branch
        # Don't intervene the ranges because counterfactual variables shouldn't be in ranges
        # intervened_ranges = tuple(
        #     variable.activate_domain_and_interventions(active_interventions) for variable in expression.ranges
        # )
        return Sum.safe(
            activate_domain_and_interventions(expression.expression, interventions, domain),
            expression.ranges,
        )
    if isinstance(expression, Fraction):
        numerator = activate_domain_and_interventions(expression.numerator, interventions, domain)
        denominator = activate_domain_and_interventions(
            expression.denominator, interventions, domain
        )
        return cast(Fraction, numerator / denominator).simplify()
    if isinstance(expression, Product):
        # TODO need full integration test to trso() function that covers this branch
        return Product.safe(
            activate_domain_and_interventions(expr, interventions, domain)
            for expr in expression.expressions
        )
    raise NotImplementedError(f"Unhandled expression type: {type(expression)}")


def all_transports_d_separated(graph, target_interventions, target_outcomes) -> bool:
    """Check if all target_interventions are d-separated from target_outcomes.

    :param graph: The graph with transport nodes in this domain.
    :param target_interventions: Set of target interventions
    :param target_outcomes: Set of target interventions
    :returns: True if all interventions are d-separated from all outcomes, False otherwise.
    """
    transportability_nodes = get_transport_nodes(graph)
    graph_without_interventions = graph.remove_in_edges(target_interventions)
    return all(
        are_d_separated(
            graph_without_interventions,
            transportability_node,
            outcome,
            conditions=target_interventions,
        )
        for transportability_node in transportability_nodes
        if transportability_node in graph_without_interventions
        for outcome in target_outcomes
    )


def trso_line9(query: TRSOQuery, district: set[Variable]) -> Expression:
    """Return the probability in the case with exactly one districts_without_interventions and it is present in districts.

    :param query: A TRSO query
    :param district: The C-component present in both districts_without_interventions and districts
    :returns: An Expression
    :raises RuntimeError: If the query's expression is zero. This should never happen
    """
    logger.debug(
        "Calling trso algorithm line 9 with expression %s \n district %s",
        query.expression,
        district,
    )
    if isinstance(query.expression, Zero):  # pragma: no cover
        # TODO if we can't create an integration test (i.e., a call to trso)
        #  that triggers this line, then it can be safely removed
        raise RuntimeError

    ordering = list(query.graphs[query.domain].topological_sort())
    ordering_set = set(ordering)
    my_product: Expression = One()
    for node in district:
        i = ordering.index(node)
        pre, post = ordering[:i], ordering[: i + 1]
        pre_set = ordering_set - set(post)
        post_set = ordering_set - set(pre)
        numerator = Sum.safe(query.expression, pre_set)
        denominator = Sum.safe(query.expression, post_set)
        my_product *= numerator / denominator
    my_product = cast(Fraction, my_product).simplify()

    logger.debug(
        "Returning trso algorithm line 9 with expression %s",
        Sum.safe(my_product, district - query.target_outcomes),
    )
    return Sum.safe(my_product, district - query.target_outcomes)


def trso_line10(
    query: TRSOQuery,
    district: set[Variable],
    new_surrogate_interventions: Dict[Population, Set[Variable]],
) -> TRSOQuery:
    """Update the TRSO query to restrict interventions and graph to district.

    :param query: A TRSO query
    :param district: The C-component of districts which contains district_without_interventions
    :param new_surrogate_interventions: Dict mapping domains to interventions performed in that domain.
    :returns: A modified TRSOQuery
    """
    ordering = list(query.graphs[query.domain].topological_sort())
    expressions = []
    for node in district:
        i = ordering.index(node)
        pre_node = set(ordering[:i])
        # note tikka splits this into two expressions that when taken together equal pre_node
        distribution = Distribution.safe(node | pre_node)
        expressions.append(
            PopulationProbability(population=query.domain, distribution=distribution)
        )

    new_query = deepcopy(query)
    new_query.target_interventions = query.target_interventions.intersection(district)
    new_query.expression = canonicalize(Product.safe(expressions))
    new_query.graphs[query.domain] = query.graphs[query.domain].subgraph(district)
    new_query.surrogate_interventions = new_surrogate_interventions
    return new_query


def trso(query: TRSOQuery) -> Optional[Expression]:  # noqa:C901
    """Run the TRSO algorithm to evaluate a transport problem.

    :param query: A TRSO query, which contains 8 instance variables needed for TRSO
    :returns: An Expression evaluating the given query, or None
    :raises RuntimeError: when an impossible condition is met
    """
    # Check that domain is in query.domains
    # check that query.surrogate_interventions keys are equals to domains
    # check that query.graphs keys are equal to domains
    logger.debug(
        "Calling trso algorithm with "
        "\t- target_interventions: %s\n"
        "\t- target_outcomes: %s\n"
        "\t- expression: %s\n"
        "\t- active_interventions: %s\n"
        "\t- domain: %s\n"
        "\t- domains: %s\n"
        "\t- graph[domain] nodes: %s\n"
        "\t- surrogate_interventions: %s",
        query.target_interventions,
        query.target_outcomes,
        query.expression,
        query.active_interventions,
        query.domain,
        query.domains,
        query.graphs[query.domain].nodes(),
        query.surrogate_interventions,
    )

    graph = query.graphs[query.domain]
    # line 1
    if not query.target_interventions:
        logger.debug("Calling trso algorithm line 1")
        return canonicalize(trso_line1(query.target_outcomes, query.expression, graph))

    # line 2
    outcome_ancestors = graph.ancestors_inclusive(query.target_outcomes)
    if get_regular_nodes(graph) - outcome_ancestors:
        new_query = trso_line2(query, outcome_ancestors)
        logger.debug("Calling trso algorithm line 2")
        return _c14n_safe(trso(new_query))

    # line 3
    additional_interventions = graph.get_no_effect_on_outcomes(
        query.target_interventions, query.target_outcomes
    )
    if additional_interventions:
        new_query = trso_line3(query, additional_interventions)
        logger.debug("Calling trso algorithm line 3")
        return _c14n_safe(trso(new_query))

    # line 4
    districts_without_interventions: set[frozenset[Variable]] = graph.remove_nodes_from(
        query.target_interventions
    ).districts()
    if len(districts_without_interventions) > 1:
        subqueries = trso_line4(
            query,
            districts_without_interventions,
        )
        terms = []
        logger.debug("Calling trso algorithm line 4 with %d subqueries", len(subqueries))
        for i, subquery in enumerate(subqueries.values()):
            logger.debug("Calling subquery %d of trso algorithm line 4", i + 1)
            term = trso(subquery)
            if term is None:
                return None
            terms.append(term)

        product = Product.safe(terms)
        summand = canonicalize(product)  # fix sort order inside product
        return canonicalize(
            Sum.safe(
                summand,
                get_regular_nodes(graph) - query.target_interventions.union(query.target_outcomes),
            )
        )

    # line 6
    if not query.active_interventions and query.surrogate_interventions:
        expressions: Dict[Population, Expression] = {}
        for domain, subquery in trso_line6(query).items():
            logger.debug("Calling trso algorithm line 6 for domain %s", domain)
            expression = trso(subquery)
            if expression is None:
                continue
            expression = activate_domain_and_interventions(
                expression, subquery.active_interventions, domain
            )
            if expression is not None:  # line7
                logger.debug(
                    "Calling trso algorithm line 7",
                )
                expressions[domain] = expression
        if len(expressions) == 1:
            return canonicalize(list(expressions.values())[0])
        elif len(expressions) > 1:
            # TODO need full integration test to trso() function that covers this branch
            #  or change to ``raise RuntimeError`` if it's not possible to reach in practice
            logger.warning("more than one expression were non-none")
            # What if more than 1 expression doesn't fail?
            # Is it non-deterministic or can we prove it will be length 1?
            return canonicalize(list(expressions.values())[0])
        else:
            # if there are no expressions, then we move on to line 8
            pass

    # line8 checks that len(districts)) != 1
    districts = graph.districts()
    # line 11 states return fail if len(districts)==1
    # keep explict tests for 0 and 1 to ensure adequate testing
    if len(districts) == 0:
        # TODO we need an integration test (i.e., call to trso()) that covers this.
        #  if it's not possible to cover in a real setting, then we can change this
        #  to raising a runtime error. Nathaniel notes that this probably only occurs
        #  if the graph is empty, but it's not clear if it makes sense to have an empty
        #  graph
        return None
    elif len(districts) == 1:
        return None
    # line 8, i.e. len(districts)>1

    # line 9
    if len(districts_without_interventions) == 0:  # pragma: no cover
        # This would happen if there is an intervention that is also an outcome,
        # which we ensure is not possible when calling the algorithm from its harness
        raise RuntimeError

    # at this point, we already checked for cases where len > 2 and len == 0,
    # so we can safely pop the only element
    district_without_interventions = districts_without_interventions.pop()
    if district_without_interventions in districts:
        return canonicalize(trso_line9(query, set(district_without_interventions)))

    # line10
    logger.debug("Calling trso algorithm line 10")
    target_districts = [
        district for district in districts if district_without_interventions.issubset(district)
    ]
    if len(target_districts) != 1:  # pragma: no cover
        # At this point, the mathematics require this, and therfore this
        # test should never evaluate to true
        raise RuntimeError
    target_district = target_districts.pop()
    # district is C' districts should be D[C'], but we chose to return set of nodes instead of subgraph
    if len(query.active_interventions) == 0:
        # TRSO Line 6 could return an empty list and skip over the returns, allowing this line to be reached.
        new_surrogate_interventions = dict()
    elif _pillow_has_transport(graph, target_district):
        return None
    else:
        new_surrogate_interventions = query.surrogate_interventions

    new_query = trso_line10(
        query,
        set(target_district),
        new_surrogate_interventions,
    )
    return _c14n_safe(trso(new_query))


def _pillow_has_transport(graph: NxMixedGraph, district: Collection[Variable]) -> bool:
    return any(is_transport_node(node) for node in graph.get_markov_pillow(district))


def check_and_raise_missing(nodes: set[Variable], graph: NxMixedGraph, name: str) -> None:
    """Verify that nodes are present in the graph.

    :param nodes: A set of nodes that should be in the graph.
    :param graph: An NxMixedGraph(), the graph of the target domain.
    :param name: Name of the set of nodes
    :raises ValueError: If any element of nodes is not in the graph.
    """
    missing_nodes = nodes - graph.nodes()
    missing_nodes_text = {node.to_text() for node in missing_nodes}
    if missing_nodes_text:
        raise ValueError(
            f"The following {name} are not in the graph: {', '.join(missing_nodes_text)}"
        )


def identify_target_outcomes(
    graph: NxMixedGraph,
    *,
    target_outcomes: Set[Variable],
    target_interventions: Set[Variable],
    surrogate_outcomes: Dict[Population, Set[Variable]],
    surrogate_interventions: Dict[Population, Set[Variable]],
) -> Expression | None:
    r"""Get the estimand for the target outcome givne the surrogate outcomes.

    .. seealso:: Originally described in https://arxiv.org/abs/1806.07172.

    :param target_outcomes: A set of target variables for causal effects.
    :param target_interventions: A set of interventions for the target domain.
    :param graph: The graph of the target domain.
    :param surrogate_outcomes: A dictionary of outcomes in other populations
    :param surrogate_interventions: A dictionary of interventions in other populations
    :returns: An Expression evaluating the given query, or None
    :raises ValueError: If the target outcomes and target interventions intersect

    The example from figure 8 of the original paper can be executed with
    the following code:

    .. code-block:: python

        from y0.algorithm.transport import identify_target_outcome
        from y0.dsl import X1, X2, Y1, Y2, Pi1, Pi2
        from y0.examples import tikka_trso_figure_8_graph

        estimand = identify_target_outcome(
            graph=tikka_trso_figure_8_graph,
            target_outcomes={Y1, Y2},
            target_interventions={X1, X2},
            surrogate_outcomes={Pi1: {Y1}, Pi2: {Y2}},
            surrogate_interventions={Pi1: {X1}, Pi2: {X2}},
        )

    This returns the following estimand:
    $\sum_{W, Z} P(W, Z) \frac{P_{X_1}^{π_1}(W, Y_1, Z)}{P_{X_1}(W, Z)}
    \frac{P_{X_2}^{π_2}(W, X_1, Y_2, Z)}{P_{X_2}(W, X_1, Z)}$
    """
    # TODO add vanilla identification check?
    # vanilla_estimand = identify_outcomes(
    #     graph=graph, outcomes=target_outcomes, treatments=target_interventions
    # )
    # if vanilla_estimand is not None:
    #     logger.warning(f"This query is identifiable without surrogates: {vanilla_estimand}")

    check_and_raise_missing(target_outcomes, graph, "target_outcomes")
    check_and_raise_missing(target_interventions, graph, "target_interventions")
    check_and_raise_missing(set().union(*surrogate_outcomes.values()), graph, "surrogate_outcomes")
    check_and_raise_missing(
        set().union(*surrogate_interventions.values()), graph, "surrogate_interventions"
    )
    outcome_is_intervention = target_outcomes.intersection(target_interventions)
    if outcome_is_intervention:
        raise ValueError(
            f"The variables {outcome_is_intervention} cannot be target_outcomes and target_interventions"
        )

    transport_query = surrogate_to_transport(
        graph=graph,
        target_outcomes=target_outcomes,
        target_interventions=target_interventions,
        surrogate_outcomes=surrogate_outcomes,
        surrogate_interventions=surrogate_interventions,
    )
    initial_expression = PopulationProbability(
        population=TARGET_DOMAIN,
        distribution=Distribution.safe(graph.nodes()),
    )
    trso_query = TRSOQuery(
        target_interventions=transport_query.target_interventions,
        target_outcomes=transport_query.target_outcomes,
        expression=initial_expression,
        active_interventions=set(),
        domain=TARGET_DOMAIN,
        domains=transport_query.domains,
        graphs=transport_query.graphs,
        surrogate_interventions=transport_query.surrogate_interventions,
    )
    return trso(trso_query)



"""Examples for smoking/cancer."""

import numpy as np
import pandas as pd

from y0.dsl import C, S, T, U, Variable

__all__ = [
    "generate_data_for_smoke_cancer",
]


def generate_data_for_smoke_cancer(
    num_samples: int,
    treatments: dict[Variable, float] | None = None,
    *,
    seed: int | None = None,
) -> pd.DataFrame:
    """Generate testing data for the smoking/cancer case study.

    :param num_samples: The number of samples to generate. Try 1000.
    :param treatments: An optional dictionary of the values to fix each variable to.
    :param seed: An optional random seed for reproducibility purposes
    :returns: A pandas Dataframe with columns corresponding
        to the variable names in the front-door back_door example
    """
    if treatments is None:
        treatments = {}
    generator = np.random.default_rng(seed)
    if U in treatments:
        u = np.full(num_samples, treatments[U])
    else:
        u = generator.normal(loc=10, scale=5, size=num_samples)

    beta0_s = 1.6
    beta_u_to_s = -0.3
    if S in treatments:
        s = np.full(num_samples, treatments[S])
    else:
        p_s = 1 / (1 + np.exp(beta0_s + u * beta_u_to_s))
        s = generator.binomial(n=1, p=p_s, size=num_samples)

    beta0_t = 0.3
    beta_u_to_t = 0.4
    beta_s_to_t = 2
    if T in treatments:
        t = np.full(num_samples, treatments[T])
    else:
        loc_t = beta0_t + u * beta_u_to_t + s * beta_s_to_t
        t = generator.normal(loc=loc_t, scale=4, size=num_samples)

    beta0_c = 1.5
    beta_s_to_c = -0.7
    beta_t_to_c = -0.1
    if C in treatments:
        c = np.full(num_samples, treatments[C])
    else:
        p_c = 1 / (1 + np.exp(beta0_c + s * beta_s_to_c + t * beta_t_to_c))
        c = generator.binomial(n=1, p=p_c, size=num_samples)

    return pd.DataFrame({S.name: s, T.name: t, C.name: c})



# -*- coding: utf-8 -*-

"""Test falsification of testable implications given a graph."""

import unittest

from y0.algorithm.conditional_independencies import get_conditional_independencies
from y0.algorithm.falsification import get_falsifications, get_graph_falsifications
from y0.examples import asia_example


class TestFalsification(unittest.TestCase):
    """Test the falsifiable implications."""

    def test_graph_falsifications(self):
        """Test the asia graph against data generated from it."""
        issues = get_graph_falsifications(
            asia_example.graph, asia_example.data, method="cressie_read"
        )
        self.assertEqual(0, len(issues.failures))
        self.assertGreater(len(issues.evidence), 0)

    def test_falsifications(self):
        """Test the asia graph against data generated from it, passing in the implications to test."""
        implications = get_conditional_independencies(asia_example.graph)
        issues = get_falsifications(implications, asia_example.data)
        self.assertEqual(0, len(issues.failures))
        self.assertEqual(len(issues.evidence), len(implications))



"""Tests for contraction functions."""

import unittest

from y0.dsl import A, B, C, D, P, Sum
from y0.mutate.contract import contract


class TestContract(unittest.TestCase):
    """Test case for contraction functions."""

    def test_contract(self):
        """Test the simple contract function."""
        self.assertEqual(P(A | B), contract(P(A, B) / P(B)))
        self.assertEqual(P(A | B), contract(P(A, B) / P(B)))
        self.assertEqual(P(A | B, C), contract(P(A, B, C) / P(B, C)))
        self.assertEqual(P(B, C | A), contract(P(A, B, C) / P(A)))
        self.assertEqual(P(A, B, C) / P(D), contract(P(A, B, C) / P(D)))

        expr = Sum[A](P(B))
        self.assertEqual(expr, contract(expr))



"""Generate example data.

This automatically works for all instances of :clas:`y0.examples.Example`
appearing in :mod:`y0.examples`. If you make a new generator, make sure
you annotate it to the correct graph and add a corresponding example
query.

This file can be run as a script directly with ``python generate.py``.
"""

import warnings
from pathlib import Path

import click
import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns
from tabulate import tabulate
from tqdm import tqdm, trange

from y0.algorithm.estimation import estimate_ace
from y0.examples import examples

warnings.simplefilter(action="ignore", category=FutureWarning)


HERE = Path(__file__).parent.resolve()


def main(seed: int = 1, num_samples: int = 1000, bootstraps: int = 500):
    """Generate example data."""
    aces = []
    for example in examples:
        if example.generate_data is None:
            continue
        directory = HERE.joinpath(example.name.lower().replace(" ", "-").replace("_", "-"))
        directory.mkdir(exist_ok=True, parents=True)

        df = example.generate_data(num_samples)
        df.to_csv(directory.joinpath("observational.tsv"), sep="\t", index=False)
        # TODO fix output
        # grid = sns.pairplot(data=df, aspect=1.65, height=1.2, plot_kws=dict(alpha=0.7))
        # grid.savefig(HERE.joinpath("napkin_observational.png"), dpi=300)

        if not example.example_queries:
            raise ValueError(f"{example.name} should have at least one example query")

        queries = [q for q in example.example_queries if 1 == len(q.treatments) == len(q.outcomes)]
        if not queries:
            raise ValueError(
                f"{example.name} should have at least one example query with a single treatment and outcome"
            )

        query = queries[0]
        treatment = list(query.treatments)[0]
        outcome = list(query.outcomes)[0]
        treatment_name = treatment.name.lower().replace("-", "").replace("_", "")

        df_treat_1 = example.generate_data(num_samples, seed=seed, treatments={treatment: 1})
        df_treat_1.to_csv(
            directory.joinpath(f"treat_{treatment_name}_1.tsv"), sep="\t", index=False
        )
        df_treat_0 = example.generate_data(num_samples, seed=seed, treatments={treatment: 0})
        df_treat_0.to_csv(
            directory.joinpath(f"treat_{treatment_name}_0.tsv"), sep="\t", index=False
        )
        actual_ace = np.mean(df_treat_1[outcome.name]) - np.mean(df_treat_0[outcome.name])
        aces.append((example.name, actual_ace))

        ace_deltas = []
        for _ in trange(bootstraps, desc=f"ACE {example.name}"):
            df = example.generate_data(num_samples)
            estimated_ace = estimate_ace(
                graph=example.graph,
                treatments=treatment,
                outcomes=outcome,
                data=df,
            )
            delta = estimated_ace - actual_ace
            ace_deltas.append(delta)

        fig, ax = plt.subplots(figsize=(4, 2.3))
        sns.histplot(ace_deltas, ax=ax)
        ax.set_title("Deviation from actual ACE")
        deviation_path = directory.joinpath("deltas.png")
        fig.savefig(deviation_path, dpi=300)
        tqdm.write(f"Wrote {example.name} deviations chart to {deviation_path}")

    click.echo(tabulate(aces, headers=["Name", "ACE"], tablefmt="github"))


if __name__ == "__main__":
    main()
